!+
! Module: emldetect_mod
!
! Authors:  Hermann Brunner - AIP
!          
! Description: Maximum likelihood EPIC source searching
! Comments: 
! 
! $Id: module.f90,v 1.5 1998/01/13 08:39:15 gvacanti Exp $
!- 
module emldetect_mod

  use param  
  use dal
  use edetect_io_ml
  use emldata
  use emlfit
  use eml_io
  use psf

  implicit none 
  
  private 

  public emldetect

contains 
  
  !---------------------------------------------------------------------
  subroutine cut_radius (x,y,scut2,ecut2,enfrac)
  !---------------------------------------------------------------------

    use cal

    real, intent(in)  :: x, y
    real, intent(out) :: scut2, ecut2
    real, dimension(max_ima), intent(out) :: enfrac 

    real, parameter :: pi = 3.14159265, sec2mm = 7500.0*pi/(180.0*3600.0)
    integer :: i,j,k,ipoint,ierr
    real    :: scut0,ecut0,ecut1,scut,ecut
    real(kind=double) :: E, theta, fraction 
    logical :: first = .true. 

    type (PsfT) :: ppsf
    type(PsfValidityRangesT) :: vrange 
    
    save

    ierr=0
    
    scut=0.
    ecut=0.
     
    if (first) then
      call CAL_setState1(instrument=EPN, &
                         AccuracyLevel=ACCURACY_MEDIUM)
      vrange = CAL_PsfValidityRanges()
      first = .false.
    end if

    do i=1,num_instr
      if (num_evlist(i).eq.0) then
        theta=sqrt((x-xcen(i))**2+(y-ycen(i))**2)*pixim 
        theta = max(min(theta,vrange%thetaMax-1.0),vrange%thetaMin+1.0)
        do j=1,num_image(i)
          ipoint = ip(i,j) 
          E=energy(ipoint)
          ppsf = Cal_getPSF(E, theta, 0d0) 
          fraction=ences
          scut0 = CAL_psfEncircledenergy(ppsf,fraction)/(pixim*sec2mm)
          scut=max(scut,scut0)
          fraction=encee
          ecut0 = CAL_psfEncircledenergy(ppsf,fraction)/(pixim*sec2mm)
          ecut=max(ecut,ecut0)
        end do
        do j=1,num_image(i)
          ipoint = ip(i,j) 
          E=energy(ipoint)
          ppsf = Cal_getPSF(E, theta, 0d0) 
          ecut1=0. 
          do k=1,99
            fraction=k*0.01
            ecut0=ecut1 
            ecut1=CAL_psfEncircledenergy(ppsf,fraction)/(pixim*sec2mm)
            if (ecut1>ecut) exit
          end do
          enfrac(j)=((k-1)*(ecut1-ecut)+k*(ecut-ecut0))*0.01/(ecut1-ecut0)
        end do
                 
      else
  
      end if

    end do   
        
    scut2 = scut**2
    ecut2 = ecut**2

  end subroutine cut_radius

  !---------------------------------------------------------------------
  subroutine sou_clus(xms,yms,scts,insrc,nsrc,scut2,xx,yy,src,in, &
                      k,nsou,indx,iwork) 
  !---------------------------------------------------------------------

    real, dimension(max_sou), intent(in)  :: xx, yy
    real, dimension(max_ima, max_sou), intent(in) :: src
    real, intent(in) :: scut2
    integer, dimension(max_sou), intent(in) :: in, indx
    integer, intent(in) :: k, nsou

    integer, dimension(max_sou), intent(inout) :: iwork

    real, dimension(max_fit), intent(out) :: xms, yms
    real, dimension(max_ima,max_fit), intent(out) :: scts   
    integer, dimension(max_fit), intent(out) :: insrc
    integer, intent(out) :: nsrc

    integer :: i,j,kk
    logical :: found      

    kk=indx(nsou-k+1)
    xms(1) = xx(kk)
    yms(1) = yy(kk)
    scts(1:nim,1)=src(1:nim,kk)
    insrc(1)=in(kk)

    nsrc=1
    found = .true.

    do while (found.and.nsrc.lt.max_sfit)
      found = .false.
      do j=k+1,nsou
        kk=indx(nsou-j+1)
        if(iwork(kk).eq.0) then
          do i=1,nsrc
            if ((xms(nsrc)-xx(kk))**2+(yms(nsrc)-yy(kk))**2 &
                .le.scut2.and.nsrc.lt.max_sfit) then
              found = .true.
              nsrc=nsrc+1
              iwork(kk)=1        
              xms(nsrc)=xx(kk)
              yms(nsrc)=yy(kk)
              scts(1:nim,nsrc)=src(1:nim,kk)
              insrc(nsrc)=in(kk)
              exit
            end if
          end do  
        endif
        if (found) exit              
      end do   
    end do
!   write(6,*)' source ',in(k),' has ',nsrc-1,' satellites ',max_sfit

    if (nsmul>1) then
      do i=nsrc,1,-1
        xms(nsmul*i)=xms(i)
        yms(nsmul*i)=yms(i)
        scts(1:nim,nsmul*i)=scts(1:nim,i)/nsmul
        insrc(2*nsmul)=insrc(i)
        do j=1,nsmul-1
          xms(nsmul*i-j)=xms(i)
          yms(nsmul*i-j)=yms(i)
          scts(1:nim,nsmul*i-j)=scts(1:nim,i)/nsmul
          insrc(2*nsmul-j)=insrc(i)
        end do
      enddo
      nsrc=nsmul*nsrc
    end if 

  end subroutine sou_clus

  !---------------------------------------------------------------------
  subroutine fill_arrs(xms,yms,nsrc,ecut2)
  !---------------------------------------------------------------------

    real, dimension(max_fit), intent(in) :: xms, yms
    real, intent(in) :: ecut2
    integer, intent(in) :: nsrc

    integer :: ierr,i,j,k,l,ipoint,ns,ne,ixs,iys,ixd,iyd,iamp, &
               x_start,x_end,y_start,y_end,ix,iy,i_offax,i_posang
    integer, dimension(max_ins) :: no_ev(max_ins)
    integer :: evmax, sum, ixmax, iymax, tempsum
    real    :: ss,ecut,xd,yd,xmax,xmin,ymax,ymin,offax,posang,x,y
    logical :: end_ev

    ierr=0

    xmax=xms(1)
    xmin=xms(1)
    ymax=yms(1)      
    ymin=yms(1)

    do k=1,nsrc
      xmax=max(xms(k),xmax)
      xmin=min(xms(k),xmin)
      ymax=max(yms(k),ymax)
      ymin=min(yms(k),ymin)        
    end do

    ecut = sqrt(ecut2)
    mask%ix_max=min(int(xmax+ecut),npix_arr(1,1))
    mask%ix_min=max(int(xmin-ecut),1)
    mask%iy_max=min(int(ymax+ecut),npix_arr(2,1))
    mask%iy_min=max(int(ymin-ecut),1)

    evmax=0
    do ix=mask%ix_min,mask%ix_max
      do iy=mask%iy_min,mask%iy_max
        mask%mask(:,ix-mask%ix_min+1,iy-mask%iy_min+1) = 0     
        do k=1,nsrc          
          if ((ix-xms(k))**2+(iy-yms(k))**2.le.ecut2) then
            mask%mask(:,ix-mask%ix_min+1,iy-mask%iy_min+1) = 1     

            if (hotfla) then 
              ! remove brightest pixel
              sum=0
              do ipoint=1,nim
                sum=sum+image(ipoint,ix,iy)
              end do
              if (sum>evmax) then
                evmax=sum
                ixmax=ix
                iymax=iy
              end if 
            end if

            do i=1,nim 
              if (detfla .and. .not.mask_image(i,ix,iy)) &
                mask%mask(i,ix-mask%ix_min+1,iy-mask%iy_min+1) = 0     
              if (expfla .and. exp_image(i,ix,iy)<=0.) &          
                mask%mask(i,ix-mask%ix_min+1,iy-mask%iy_min+1) = 0     
            end do
          end if
        end do
      end do
    end do 
    if (hotfla) mask%mask(:,ixmax-mask%ix_min+1,iymax-mask%iy_min+1) = 0


    l = 1
    do i=1,num_instr
        
      if (num_evlist(i).eq.0) then

        do j=1,num_image(i) 
          ipoint=ip(i,j)
          do iy=mask%iy_min,mask%iy_max      
            do ix=mask%ix_min,mask%ix_max
              if (mask%mask(ipoint,ix-mask%ix_min+1, &
                            iy-mask%iy_min+1).eq.1) then 
                event(l)%n=image(ipoint,ix,iy)
                event(l)%ipt=ipoint
                l=l+1
              end if 
            end do
          end do
        end do

      else

! read in event lists
        ns=ip(i,1)
        ne=ip(i,num_image(i))

! open events file and get list of small maps
!cc       call pix2sky(float(ix_min)
!cc       call pix2sky(
        call get_smaps (no_ev(i),x_start,y_start,x_end,y_end)

! go through list of small maps and read one event at a time
        do while (.not.end_ev) 
          call get_event (no_ev(i),ixs,iys,ixd,iyd,iamp,end_ev)
          call sky2pix(ixs,iys,x,y,start_arr(1,i),step_arr(1,i))
          ix=nint(x)
          iy=nint(y)
          if (ix.ge.mask%ix_min.and.ix.le.mask%ix_max.and. &
              iy.ge.mask%iy_max.and.iy.le.mask%iy_max.and. &
                mask%mask(1,ix-mask%ix_min+1,iy-mask%iy_min+1).eq.1) &
          then
            event(l)%xs  = x
            event(l)%ys  = y 
            event(l)%n   = 1
            event(l)%ins = i
            do j=ns,ne
              if (iamp.ge.ampl_low(j).and.iamp.le.ampl_high(j)) &
                event(l)%ipt = j
            end do 
            xd=ixd-xdcen(i)
            yd=iyd-ydcen(i)
            offax = sqrt(xd**2+yd**2)*detpix(i)/60.
            i_offax = min(int(offax/epsmax(i)*max_offax)+1,max_offax)
            event(l)%offax  = i_offax
            posang = atan2(xd,yd)
            i_posang = min(int(posang/twopi*max_posang)+1,max_posang)
            event(l)%posang = i_posang
            hist(i)%hist(i_offax,i_posang)=hist(i)%hist(i_offax,i_posang)+1.
            l = l+1 
          end if
        end do

! accumulate off axis & position angle histogram        
        hist(i)%ns_offax=max_offax  
        hist(i)%ne_offax=1  
        hist(i)%ns_posang=max_posang  
        hist(i)%ne_posang=1  

        ss=0.
        do i_offax=1,max_offax
          do i_posang=1,max_posang
            if (hist(i)%hist(i_offax,i_posang).gt.0.) then
              hist(i)%ns_offax=min(hist(i)%ns_offax,i_offax)
              hist(i)%ne_offax=max(hist(i)%ns_offax,i_offax)
              hist(i)%ns_posang=min(hist(i)%ns_posang,i_posang)
              hist(i)%ne_posang=max(hist(i)%ns_posang,i_posang)
              ss=ss+hist(i)%hist(i_offax,i_posang)
            end if
          end do
        end do

        hist(i)%hist(:,:) = hist(i)%hist(:,:)/ss       

      end if       

    end do

    npts = l-1

  end subroutine fill_arrs

  !---------------------------------------------------------------------
  subroutine sky2pix (xsky,ysky,x,y,start,step)
  !---------------------------------------------------------------------

    integer          xsky,ysky
    real             x,y
    double precision start(2),step(2)

    x=(xsky-start(1)+0.5)/step(1)
    y=(ysky-start(2)+0.5)/step(2)

  end subroutine sky2pix

  !---------------------------------------------------------------------
  subroutine pix2sky(x,y,xsky,ysky,start,step)
  !---------------------------------------------------------------------

    real              x,y
    integer           xsky,ysky
    double precision  start(1),step(1)

    xsky=nint(x)
    ysky=nint(y)

  end subroutine pix2sky

  !---------------------------------------------------------------------
  subroutine rd_data (ierr)
  !---------------------------------------------------------------------
 
    integer           i,j,ipoint,maxdim,ierr
    parameter         (maxdim=2)
    integer           dattyp,filtyp,iomode, ipxid, ix, iy
    integer, dimension(max_ins) :: num_ima0
    real              exposure
    real(kind=single), dimension(3) :: xnull, ynull
    double precision  start(maxdim),step(maxdim)
    character*1024    fname

    dattyp   = 10
    filtyp   = 1
    iomode   = 0
    xnull = 0.
    ynull = 0.
    xid_arr=-1
    crpix1=-2d30
    crpix2=-2d30
    crval1=-2d30
    crval2=-2d30
    crdelt1=-2d30
    crdelt2=-2d30

    nim = ip(num_instr,num_image(num_instr))

! add XID  band to pointer array ip if at least band 1-3 are present
    num_ima0 = num_image
    do i = 1,num_instr
      j=expip(i)
      if (num_image(i) >= xiddef(2,j) .and. xidfla) then
        num_image(i)=num_image(i)+1
        nim=nim+1
        ip(i,num_image(i)) = nim
        xid_arr(i)=num_image(i)
        pimin(nim)=pimin(ip(i,xiddef(1,j)))
        pimax(nim)=pimax(ip(i,xiddef(2,j)))
      end if
    end do

    do i = 1,num_instr
      do j = 1,num_ima0(i)
        ipoint=ip(i,j)  

        fname = c_image(ip(i,j))

        call co_image(fname,start,step,npix,exposure,xcen(i),ycen(i), &
                      ra_pnt,dec_pnt,pa_pnt,datetime, &
                      crpix1,crpix2,crval1,crval2,crdelt1,crdelt2, &
                      crota2,xnull,ynull,expip(i),ipoint,nim)

        pixim = 3600.0 * dabs(crdelt1) 

        if (j.eq.1) then
          start_arr(:,i) = start
          step_arr (:,i) = step 
          npix_arr (:,i) = npix 
        end if 

        fname = c_bgima(ip(i,j))

        call co_bgima(fname,npix,crpix1,crpix2,crval1,crval2,crdelt1,crdelt2, &
                      xnull,ynull,expip(i),ipoint,nim)

        if (expfla) then
           
          fname = c_expima(ip(i,j))
          call co_expima(fname,npix,crpix1,crpix2,crval1,crval2,crdelt1,crdelt2, &
                         xnull,ynull,expip(i),ipoint, &
                         xcen(i),ycen(i),exp_center(ipoint),nim)

          !exp_center(ipoint)=exp_image(ipoint,int(xcen(i)),int(ycen(i)))
        else
          if (i==1.and.j==1) allocate (exp_image(nim,npix(1),npix(2)))
          exp_image=1.0          
        end if

      end do

      if (detfla) then
        fname = c_detima(i)
        call co_mask(fname,npix,crpix1,crpix2,crval1,crval2,crdelt1,crdelt2, &
                     xnull,ynull,expip(i),ip(i,1),nim)
        
        do j = 2,num_image(i)
          do iy = 1,npix(2)
            do ix = 1,npix(1)
              mask_image(ip(i,j),ix,iy)=mask_image(ip(i,1),ix,iy)
            end do
          end do
        end do
      else
        if (i==1) allocate (mask_image(nim,npix(1),npix(2)))
        mask_image=.true.
      end if   
        
    end do

    if (crpix1<-1d30) crpix1=0d0
    if (crpix2<-1d30) crpix2=0d0
    if (crval1<-1d30) crval1=0d0
    if (crval2<-1d30) crval2=0d0
    if (crdelt1<-1d30) crdelt1=0d0
    if (crdelt2<-1d30) crdelt2=0d0

! add XID band images if band 1 - 3 are available

    do i = 1,num_instr
     j=expip(i)
     if (num_image(i) >= xiddef(2,j) .and. xidfla) then 
      ipxid=ip(i,num_image(i))
      do iy=1,npix(2)
        do ix=1,npix(1)
          image(ipxid,ix,iy)   = image(ip(i,xiddef(1,j)),ix,iy) + &
                                 image(ip(i,xiddef(2,j)),ix,iy)
          bg_image(ipxid,ix,iy) = bg_image(ip(i,xiddef(1,j)),ix,iy) + &
                                  bg_image(ip(i,xiddef(2,j)),ix,iy)
          exp_image(ipxid,ix,iy) = (exp_image(ip(i,xiddef(1,j)),ix,iy) + &
                                    exp_image(ip(i,xiddef(2,j)),ix,iy)) / 2.
          mask_image(ipxid,ix,iy) = mask_image(ip(i,xiddef(1,j)),ix,iy) .and. &
                                    mask_image(ip(i,xiddef(2,j)),ix,iy)
        end do
      end do
      exp_center(ipxid)=(exp_center(ip(i,xiddef(1,j)))+ &
                         exp_center(ip(i,xiddef(2,j))))/2.0
      energy(ipxid)=(energy(ip(i,xiddef(1,j)))+energy(ip(i,xiddef(2,j))))/2.0 
      ecf(ipxid)=xidecf(i)
     end if
    end do

    allocate (bg_image0(nim,npix(1),npix(2)))
    bg_image0 = bg_image       
 
  end subroutine rd_data

  !---------------------------------------------------------------------
  subroutine sorind(a,w,ind,indw,n,nw,ierr)
  !---------------------------------------------------------------------

    integer :: n,nw,ierr,ind(n),indw(nw),i,j
    real :: a(n),w(nw)

    ierr=0

! index the array ind such that a is in ascending order
    do i = 1,n
      ind(i) = 0
    enddo
    do i = 1,nw
      indw(i) = 0
    enddo

    if (n.gt.1) then
      call indexr(n,a,ind)
    else if (n.eq.1) then
      ind(1) = 1
    else if (n.eq.0) then
      ind(1) = 0
    endif

! sort a according to ind
    do i = 1,n
      w(i) = a(ind(i))
    enddo
    do j = 1,n
      a(j) = w(j)
    enddo
    do j = 1,n
      indw(j) = ind(j)
    enddo

  end subroutine sorind    


  !---------------------------------------------------------------------
  subroutine emldetect
  !---------------------------------------------------------------------

!  author : GRH              date: 15-APR-1987 15:48
!  update : GRH              date: 3-DEC-1992 09:13  
!  update : PTK              date: 05-JAN-1993 17:27 EXSAS implementation
!  update : PTK              date: 14-JUN-1993 20:17 table output / correct
!                                                    bugs
!  update : EXSAS            date: 15-OCT-1993 08:52
!  update : PTK              date: 08-NOV-1993 10:41 POSFLA, new parameter
!                                                    IPOS in CGAUSN
!  update : PTK              date: 18-JAN-1994 09:55 NELEM1,NELEM2
!  update : EXSAS            date: 19-JAN-1994 16:43
!  update : HEB              date: 27-MAY-1996

!     character*80  table
!     character*2   io
!     character*70  d2par

      integer ::    k, ll, ierr, &
                    indx(max_sou),insrc(max_fit), &
                    iwork(max_sou),kk,nsou, &
                    in(max_sou), nsrc, istat, id_clu 
      real ::       scut2, ecut2, xx(max_sou), yy(max_sou)
      real(kind=single), dimension(max_fit) :: xms, yms, exms, eyms, eradec 
      real(kind=single), dimension(max_ima, max_fit) :: scts, escts, rate, erate
      real(kind=single), dimension(max_ima, max_sou) :: src
      real(kind=single), dimension(max_sou) :: sn, rwork
      real(kind=double), dimension(max_fit) :: ra, dec, lii, bii
      integer, dimension(3,max_fit) :: chipix,chipiy,ccdinr 
 
      real, dimension(max_ima,max_fit) :: exi_ml 
      real, dimension(max_fit) :: ext, ext_err, ext_ml 
      real, dimension (max_ima) :: enfrac
      character(len=5) :: flags
                    
      call mulmlpari

! Read input images
      call rd_data (ierr)

      call get_sources (nsou,xx,yy,sn,src,in)

! Sort the array, key is SN
      call sorind(sn, rwork, indx, iwork, nsou, max_sou, istat)

      iwork=0
      id_clu=0
      ll=0
       
! Do loop, start from brightest source 
      do k=1,nsou
!             write(*,*) k, " sources processed"
        if (mod(k,10) == 0) then
           write(errstr,*) k, " sources processed"
           call message(AppMsg,Verbose)
        end if
        kk=indx(nsou-k+1)
        if (iwork(kk).eq.0) then
          iwork(kk)=1
          id_clu=id_clu+1

! Calculate cut radii
          call cut_radius (xx(k),yy(k),scut2,ecut2,enfrac)

! Identify source cluster
          call sou_clus (xms,yms,scts,insrc,nsrc,scut2,xx,yy,src,in, &
                         k,nsou,indx,iwork) 
         
! Fill data arrays
          call fill_arrs (xms,yms,nsrc,ecut2)

          if (npts.gt.nim+4) then

! Perform multi-source ML fit
            call mulsou(nsrc,xms,yms,ext,scts,rate,ra,dec,lii,bii, &
                        chipix,chipiy,ccdinr, &
                        exms,eyms,ext_err,escts,erate,eradec, &
                        exi_ml,ext_ml,ecut2)
! Set dummy output values
            flags="     "
           
! write source info to table
            call wr_stab(ll,nsrc,insrc,id_clu,xms,yms,scts,rate,ra,dec,lii,bii,&
                         chipix,chipiy,ccdinr,&
                         exms,eyms,escts,erate,eradec,exi_ml,ext,ext_err,&
                         ext_ml,enfrac,ecut2,flags)
          end if 

        endif

      end do

! Close output table       
      ! ll = -ll
      call mlout(0,ll,k,0,0,0,0,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0., &
                 0.,0.,0.,0.,0d0,0d0,0.,0d0,0d0,0,0,0,0.,0.,0.,0.,0.,0.,0.,0., &
                 0.,0.,flags) 


! write out source map
      if (smpfla) call wr_srcmap

    end subroutine emldetect
  
end module emldetect_mod








