module emlfit

  use param
  use dal
  use sla
  use edetect_io_ml
  use edetect_gen
  use emldata
  use psf


  implicit none

  private

  public mulsou, model, mlfit, funct, like, ordr, pordr

contains

  !---------------------------------------------------------------------
      subroutine mulsou (nsrc,xms,yms,sigma,scts,rate,ra,dec,lii,bii, &
                         chipix,chipiy,ccdinr, &
                         exms,eyms,esigma,escts,erate,eradec, &
                         exi_ml,ext_ml,ecut2)
  !---------------------------------------------------------------------

      real, intent(in) :: ecut2
      integer, intent(inout) :: nsrc
      real, dimension(max_fit), intent(inout) :: xms, yms, sigma, &
                                                 exms, eyms, esigma, ext_ml
      real(kind=double), dimension(max_fit), intent(out) :: ra, dec, lii, bii
      real(kind=single), dimension(max_fit) ::  eradec
      real, dimension(max_ima,max_fit), intent(inout) :: scts, escts, &
                                                         rate, erate,exi_ml
      integer, dimension(3,max_fit), intent(out) :: chipix,chipiy,ccdinr 

      real ::       yfit1(max_events),yfit2(max_ima),a(max_par), &
                    deltaa(max_par),sigmaa(max_par),afit(max_par), &
                    delfit(max_par),sigfit(max_par),yfit3,dml(max_fit), &
                    fe(max_par),amerk(max_par),like0,lik_arr(max_ima), &
                    likmerk,lik_amerk(max_ima),likold,df, &
                    felhi,fello, flamda, sc,rcal, x1,y1,exp,expn,expj,ene
      real(kind=double) :: xr,yr,dra,ddec,r0r,d0r,rr,dr,sin_pa,cos_pa, &
                    lr,br,fac
      integer ::    ierr,ioff0,ioff1,ioff,i,ifel,ix0,iy0, &
                    npsou,ipoint,iordr(max_par),iorde(max_par), &
                    j,insrc(max_fit),ical, &
                    ix,iy,l,m,n,nlauf,ntermf,ntfix,nffix,np,nima,nband,masksum
      logical       done
      character(len=6) :: minmeth 

      integer, parameter :: irad = 100, maxfun=1000, iprint=0
      real, parameter :: delt = 1e-3
       real(kind=double), parameter ::  rhobeg=0.1, rhoend=1.D-6
  
      minmeth='mrqmin'
!      minmeth='powell' 
 
!     write(*,*) " ************* start MULSOU ***********"
!     write(*,*) "x,y:", xms,yms

      ierr=0
      ntfix = 0
      nlauf=0
      ical=45/pixim
      rcal=ical

      nspar=3
      npsou=nim+nspar
      np=npsou*nsrc
      do i=1,nsrc
        ioff=npsou*(i-1)
        sc=0.0
        do j=1,nim
          expj=exp_image(j,max(1,min(npix(1),nint(xms(i)))), &
                           max(1,min(npix(2),nint(yms(i)))))
          expj=max(expj,1.0)
          if (scts(j,i)<=0.) then 
            scts(j,i)=100./expj
          else if (.not. scts(j,i) > 0.) then
!           catch NULL values from input list           
            scts(j,i)=0.0
          endif
            a(ioff+3+j)=scts(j,i) 
            deltaa(ioff+3+j)=max(sqrt(scts(j,i)*expj)/expj,1.E-5)
            sc=sc+a(ioff+3+j)*expj
        end do
!       assume 10 counts, if all input values were 0 or NULL:  
        sc=max(sc,10.)
        a(ioff+1)=xms(i)
        deltaa(ioff+1)=2.0/sqrt(sc)
        a(ioff+2)=yms(i)
        deltaa(ioff+2)=2.0/sqrt(sc)
        if (extfla) then 
          a(ioff+3)=5.0
        else
          a(ioff+3)=0.
        endif
        deltaa(ioff+3)=1.0 
      end do



    done = .false.
    do while (.not. done)

!-----------------------------------------------------------
      iordr(1)=0
      call ordr(np,np,iordr)                                           



      nspar=3
      ntermf=np

! optionally fix sigmas in fit
      if (.not. extfla) then
        do i=1,nsrc
          ioff=npsou*(i-1)
          call ordr(np,ioff+3,iordr)
        end do
        ntermf=ntermf-nsrc
        ntfix=nsrc
        nspar=nspar-1
      end if


! optionally fix positions in fit
      if (.not. posfla) then
        do n=1,nsrc
          ioff=npsou*(n-1)
          call ordr(ntermf,ioff+1,iordr)
          call ordr(ntermf,ioff+2,iordr)
        enddo
        ntermf=ntermf-2*nsrc
        ntfix=ntfix+2*nsrc
        nspar=nspar-2
      endif

! fix count rate at zero if mask in entire source region is zero:
 
      do n=1,max_fit
        do j=1,max_ima
          maskzero(n,j)=0
        end do
      end do
     
      do n=1,nsrc
        ioff=npsou*(n-1)
        do j=1,nim
          masksum=0
          do ix=mask%ix_min,mask%ix_max
            do iy=mask%iy_min,mask%iy_max
              if ((ix-xms(n))**2+(iy-yms(n))**2.le.ecut2) then
                masksum=masksum+mask%mask(j,ix-mask%ix_min+1,iy-mask%iy_min+1)
              endif
            end do
          end do
          if (masksum .eq. 0) then 
            a(ioff+3+j)=0.0
            deltaa(ioff+3+j)=0.0
            call ordr(ntermf,ioff+3+j,iordr)
            ntermf=ntermf-1
            ntfix=ntfix+1
            maskzero(n,j)=1
          endif
        end do
      end do

!-------------------------------------------

      call pordr(np,a,afit,iordr,1)
      call pordr(np,deltaa,delfit,iordr,1)
      call pordr(np,sigmaa,sigfit,iordr,1)
      if (minmeth .eq. 'powell') then 
        call powmin(ntermf, rhobeg, rhoend, iprint, maxfun, afit, iordr, xms, yms, &
                  ecut2, yfit1, like0, lik_arr, nsrc, ntfix, ierr)  
!        write(*,*)" a) ",like0                       
      else
        count_model=0       
        do nlauf=1,50
          flamda=1e-2    
          call mlfit(ntermf,afit,delfit,sigfit,iordr,xms,yms,ecut2, &
                                  flamda,yfit1,like0,lik_arr, nsrc, ntfix, ierr)
!        write(*,*)" a) ",like0   
          if(abs(likold-like0).le.delt) then
!            write(*,*) "nlauf:", nlauf
            exit
          endif
          likold=like0                                                    
        end do
!        write(*,*) "# convolutions:", count_model
      endif  
      call pordr(np,a,afit,iordr,-1)                                   
      call pordr(np,deltaa,delfit,iordr,-1)                            
      call pordr(np,sigmaa,sigfit,iordr,-1)   


      df=npts-np                                                       

! save parameters ......
      likmerk=like0
      lik_amerk=lik_arr
      do n=1,np
        amerk(n)=a(n)
      enddo

! calculate extent likelihood
      if (extfla) then
        do n=nsrc,1,-1
          ioff=npsou*(n-1)
!!!          if (nsrc == 1) then
!!!            a(3)=0.
!!!            call pordr(np,a,afit,iordr,1)
!!!            call funct(afit,iordr,yfit1,yfit2,yfit3,nsrc,np,xms,yms,ecut2)
!!!            call pordr(np,a,afit,iordr,-1)
!!!            call like(like0,lik_arr,yfit1,yfit2)            
!!!          else
! zero extent of relevant component
            a(ioff+3)=0. 

            iorde=iordr
            call ordr(ntermf,ioff+3,iorde)
            nffix=1

             call pordr(np,a,afit,iorde,1)
             call pordr(np,deltaa,delfit,iorde,1)
             call pordr(np,sigmaa,sigfit,iorde,1)
             if (minmeth .eq. 'powell') then 
               call powmin(ntermf-nffix, rhobeg, rhoend, iprint, maxfun, afit, iordr, xms, yms, &
                         ecut2, yfit1, like0, lik_arr, nsrc, ntfix+nffix, ierr) 
!               write(*,*)" b) ",like0           
             else
               do nlauf=1,50
                 flamda=1e-2      
                  call mlfit(ntermf-nffix,afit,delfit,sigfit,iorde,xms,yms, &
                          ecut2,flamda,yfit1,like0,lik_arr,nsrc,ntfix+nffix,ierr)
!               write(*,*)" b) ",like0  
!          write(*,*) "b) nlauf, likold-like0, delt:",nlauf,likold-like0, delt  
                 if(abs(likold-like0).le.delt) exit                   
                 likold=like0                                                    
               end do
             endif  
             call pordr(np,a,afit,iorde,-1)                                   
             call pordr(np,deltaa,delfit,iorde,-1)                            
             call pordr(np,sigmaa,sigfit,iorde,-1)  
 
                        
!!!          end if
          ext_ml(n)=like0-likmerk
!  correct also extent likelihood for number of free parameters
!  (28/02/2003 GL) 
!  (changed 16/03/2003: point source -> extended is one additional parameter)
          ext_ml(n)=gammq(0.5,ext_ml(n))
! store point source fit if extent is not significant
          if (ext_ml(n) < dmlext .or. amerk(ioff+3) < 0.5) then
            ext_ml(n)=0.
            likmerk=like0
            do j=1,nim+3
              amerk(ioff+j)=a(ioff+j)
            end do
          end if  
        end do
      end if

! now calculate the likelihood of each of the sources
      do n=nsrc,1,-1
        ioff=npsou*(n-1)

        if(nsrc.eq.1)then

! simply calculate likelihood for one source only
          do j=1,nim
            a(3+j)=0.
          end do
          call pordr(np,a,afit,iordr,1)
          call funct(afit,iordr,yfit1,yfit2,yfit3,nsrc,np,xms,yms,ecut2)
          call pordr(np,a,afit,iordr,-1)
          call like(like0,lik_arr,yfit1,yfit2)
        else


 
! zero flux of relevant component
          nffix=0
          if (posfla) then
            call ordr(ntermf,ioff+1,iordr)
            call ordr(ntermf,ioff+2,iordr)
              nffix=nffix+2
          end if
          if (extfla .and. ext_ml(n) >= dmlext) then
            call ordr(ntermf,ioff+3,iordr)
            nffix=nffix+1
          end if
          do j=1,nim
            a(ioff+3+j)=0.
            call ordr(ntermf,ioff+3+j,iordr)
            nffix=nffix+1
          end do
      
! fix parameters 

          call pordr(np,a,afit,iordr,1)
          call pordr(np,deltaa,delfit,iordr,1)
          call pordr(np,sigmaa,sigfit,iordr,1)
          if (minmeth .eq. 'powell') then 
            call powmin(ntermf-nffix, rhobeg, rhoend, iprint, maxfun, afit, iordr, xms, yms, &
                        ecut2, yfit1, like0, lik_arr, nsrc, ntfix+nffix, ierr) 
!             write(*,*)" c) ",like0             
          else
            do nlauf=1,50
              flamda=1e-2      
              call mlfit(ntermf-nffix,afit,delfit,sigfit,iordr,xms,yms, &
                         ecut2,flamda,yfit1,like0,lik_arr,nsrc,ntfix+nffix,ierr)
!             write(*,*)" c) ",like0 
              if(abs(likold-like0).le.delt) exit                   
              likold=like0                                                    
            end do
          endif  
          call pordr(np,a,afit,iordr,-1)
          call pordr(np,deltaa,delfit,iordr,-1)
          call pordr(np,sigmaa,sigfit,iordr,-1)                            
        endif

        !dml(n)=like0-likmerk
        do j=1,nim 
          exi_ml(j,n)=lik_arr(j)-lik_amerk(j)
        end do
        fac=1.0
        nband=0
        dml(n)=0.         
        do j=1,num_instr
          nima=num_image(j)
          if (xid_arr(j)>0) nima=xid_arr(j)-1
          do i=1,nima
            ipoint=ip(j,i)
            dml(n)=dml(n)+exi_ml(ipoint,n)
            nband=nband+1
            fac=fac*nband
          end do
        end do
        dml(n)=gammq(real(nspar+nband)/2.,dml(n))
        do j=1,nim 
          exi_ml(j,n)=max(lik_arr(j)-lik_amerk(j),0.)
        end do
        
        if(dml(n).lt.dmlmin)then
! throw out component
          do l=n+1,nsrc
            ioff0=npsou*(l-2)
            ioff1=npsou*(l-1)
            do j=1,3
              a(ioff0+j)=a(ioff1+j)
            end do          
            do j=1,nim
              a(ioff0+3+j)=a(ioff1+3+j)
              exi_ml(j,l-1)=exi_ml(j,l)
            enddo
            dml(l-1)=dml(l)
            ext_ml(l-1)=ext_ml(l)
            insrc(l-1)=insrc(l)
          enddo
          nsrc=nsrc-1
          if(nsrc.eq.0)then
            return
          endif
          np=nsrc*npsou
          exit
        else  

! get back best-fit parameters
          do m=1,np
            a(m)=amerk(m)
          enddo
          like0=likmerk
          done = .true.
        endif
      enddo
     end do 
  
! error loop begins   
      if (errfla) then

        iordr(1)=0
        call ordr(np,np,iordr)                                           
 
! now position errors for position fixing
!       npsou=nim+nspar
        do i=1,nsrc
          do j=1,npsou
            if (j <= 2 .and. posfla .or. &
                j == 3 .and. extfla .and. ext_ml(i)>=dmlext .or. &
                j >= 4) then

              ifel=npsou*(i-1)+j

              call ordr(np,ifel,iordr)

              flamda=1.0e-2
              call pordr(np,a,afit,iordr,1)
              call pordr(np,deltaa,delfit,iordr,1)
              call pordr(np,sigmaa,sigfit,iordr,1)
              call felsu(np,afit,delfit,sigfit,iordr,xms,yms, &
                         flamda,nsrc,like0,fello,felhi,np,ecut2)
              call pordr(np,a,afit,iordr,-1)
              call pordr(np,deltaa,delfit,iordr,-1)
              call pordr(np,sigmaa,sigfit,iordr,-1)
              if (fello .le. 0) then
                ! if lower error is set to 0, use felhi as error
                fe(ifel)=max(felhi,0.)
              else                
                fe(ifel)=0.5*(max(fello,0.)+max(felhi,0.))
              endif
            end if
          end do
        end do
      
      else
        fe = 0.
      end if
                                                                      
! assign output values
      do i=1,nsrc
        ioff=npsou*(i-1)
        xms(i)=a(ioff+1)
        exms(i)=fe(ioff+1)
        yms(i)=a(ioff+2)
        eyms(i)=fe(ioff+2)
        sigma(i)=a(ioff+3)
        if (sigma(i) .gt. 0.) then
          esigma(i)=fe(ioff+3)
        else
          esigma(i)=0.
        end if 
        do j=1,nim
          scts(j,i)=a(ioff+3+j)
          escts(j,i)=fe(ioff+3+j)
        end do

        cos_pa = dcos(crota2 * pi/180d0)
        sin_pa = dsin(crota2 * pi/180d0)
        xr = (xms(i)-crpix1) * crdelt1 * pi/180d0
        yr = (yms(i)-crpix2) * crdelt2 * pi/180d0
        dra =  xr * cos_pa - yr * sin_pa
        ddec = xr * sin_pa + yr * cos_pa
        r0r = crval1 * pi/180d0
        d0r = crval2 * pi/180d0
        call antisp(dra,ddec,r0r,d0r,rr,dr)
        ra(i) = rr * 180d0/pi
        dec(i) = dr *180d0/pi
        eradec(i)  = sqrt(exms(i)**2+eyms(i)**2) * pixim

        call sla_EQGAL(rr,dr,lr,br)
        lii(i) = lr * 180d0/pi
        bii(i) = br * 180d0/pi

      enddo
! get raw coordinates
      call rawcoords(ra,dec,nsrc,ra_pnt,dec_pnt,pa_pnt,datetime, &
                     chipix, chipiy, ccdinr) 
      
! update background map
      l=1
      do i=1,num_instr
        do j=1,num_image(i)
          ipoint=ip(i,j)
          ene=energy(ipoint)
          do n=1,nsrc
            ix0=xms(n)
            iy0=yms(n)
            exp=0.
            expn=0.            
            do ix=ix0-irad,ix0+irad
              do iy=iy0-irad,iy0+irad
               if ((ix-ix0)**2+(iy-iy0)**2<irad**2) then
                if (ix>=1.and.ix<=npix(1).and.iy>=1.and.iy<=npix(2)) then
                  x1=ix
                  y1=iy 
                  if ((x1-xms(n))**2+(y1-yms(n))**2<ical**2) then
!                    if (point_resp(i) .ge. 1 .and. point_resp(i) .le. 3) then
!                       bg_image(ipoint,ix,iy) = bg_image(ipoint,ix,iy) + &
!                         scts(ipoint,n) * &
!                         psf1(point_resp(i),energy,j,ix,iy, &
!                         xms(n)-xoff(expip(i)),yms(n)-yoff(expip(i)), &
!                         xcen(i),ycen(i),sigma(n),rcal) * &
!                         exp_image(ipoint,ix,iy)
!                       exp = exp + & 
!                         psf1(point_resp(i),energy,j,ix,iy, &
!                         xms(n)-xoff(expip(i)),yms(n)-yoff(expip(i)), &
!                         xcen(i),ycen(i),sigma(n),rcal) * &
!                         exp_image(ipoint,ix,iy)
!                       expn = expn + &
!                         psf1(point_resp(i),energy,j,ix,iy, &
!                         xms(n)-xoff(expip(i)),yms(n)-yoff(expip(i)), &
!                         xcen(i),ycen(i),sigma(n),rcal)
!                     else if (point_resp(i) .eq. 4) then 
!                       bg_image(ipoint,ix,iy) = bg_image(ipoint,ix,iy) + &
!                         scts(ipoint,n)*psf1(1,energy,j,ix,iy, &
!                         xms(n)-xoff(expip(i)),yms(n)-yoff(expip(i)), &
!                         xcen(i),ycen(i),sigma(n),rcal) * &
!                         exp_image(ipoint,ix,iy)
!                       exp = exp + & 
!                         psf1(1,energy,j,ix,iy, &
!                         xms(n)-xoff(expip(i)),yms(n)-yoff(expip(i)), &
!                         xcen(i),ycen(i),sigma(n),rcal) * &
!                         exp_image(ipoint,ix,iy)
!                       expn = expn + &
!                         psf1(1,energy,j,ix,iy, &
!                         xms(n)-xoff(expip(i)),yms(n)-yoff(expip(i)), &
!                         xcen(i),ycen(i),sigma(n),rcal)
!                     else
!                        call error("WrongPSF","Wrong PSF number") 
!                     endif 
                    select case (point_resp(i))       
                      case(1)
                       bg_image(ipoint,ix,iy) = bg_image(ipoint,ix,iy) + &
                         scts(ipoint,n) * &
                         psf1(energy,j,ix,iy, &
                         xms(n)-xoff(expip(i)),yms(n)-yoff(expip(i)), &
                         xcen(i),ycen(i),sigma(n),rcal) * &
                         exp_image(ipoint,ix,iy)
                       exp = exp + & 
                         psf1(energy,j,ix,iy, &
                         xms(n)-xoff(expip(i)),yms(n)-yoff(expip(i)), &
                         xcen(i),ycen(i),sigma(n),rcal) * &
                         exp_image(ipoint,ix,iy)
                       expn = expn + &
                         psf1(energy,j,ix,iy, &
                         xms(n)-xoff(expip(i)),yms(n)-yoff(expip(i)), &
                         xcen(i),ycen(i),sigma(n),rcal)
                      case(2)
                       bg_image(ipoint,ix,iy) = bg_image(ipoint,ix,iy) + &
                         scts(ipoint,n)*psf2(energy,j,ix,iy, &
                         xms(n)-xoff(expip(i)),yms(n)-yoff(expip(i)), &
                         xcen(i),ycen(i),sigma(n),rcal) * &
                         exp_image(ipoint,ix,iy)
                       exp = exp + & 
                         psf2(energy,j,ix,iy, &
                         xms(n)-xoff(expip(i)),yms(n)-yoff(expip(i)), &
                         xcen(i),ycen(i),sigma(n),rcal) * &
                         exp_image(ipoint,ix,iy)
                       expn = expn + &
                         psf2(energy,j,ix,iy, &
                         xms(n)-xoff(expip(i)),yms(n)-yoff(expip(i)), &
                         xcen(i),ycen(i),sigma(n),rcal)
                      case(3)
                       bg_image(ipoint,ix,iy) = bg_image(ipoint,ix,iy) + &
                         scts(ipoint,n)*psf3(energy,j,ix,iy, &
                         xms(n)-xoff(expip(i)),yms(n)-yoff(expip(i)), &
                         xcen(i),ycen(i),sigma(n),rcal) * &
                         exp_image(ipoint,ix,iy)
                       exp = exp + & 
                         psf3(energy,j,ix,iy, &
                         xms(n)-xoff(expip(i)),yms(n)-yoff(expip(i)), &
                         xcen(i),ycen(i),sigma(n),rcal) * &
                         exp_image(ipoint,ix,iy)
                       expn = expn + &
                         psf3(energy,j,ix,iy, &
                         xms(n)-xoff(expip(i)),yms(n)-yoff(expip(i)), &
                         xcen(i),ycen(i),sigma(n),rcal)
                      case(4)
                       bg_image(ipoint,ix,iy) = bg_image(ipoint,ix,iy) + &
                         scts(ipoint,n)*psf1(energy,j,ix,iy, &
                         xms(n)-xoff(expip(i)),yms(n)-yoff(expip(i)), &
                         xcen(i),ycen(i),sigma(n),rcal) * &
                         exp_image(ipoint,ix,iy)
                       exp = exp + & 
                         psf1(energy,j,ix,iy, &
                         xms(n)-xoff(expip(i)),yms(n)-yoff(expip(i)), &
                         xcen(i),ycen(i),sigma(n),rcal) * &
                         exp_image(ipoint,ix,iy)
                       expn = expn + &
                         psf1(energy,j,ix,iy, &
                         xms(n)-xoff(expip(i)),yms(n)-yoff(expip(i)), &
                         xcen(i),ycen(i),sigma(n),rcal)
                      case default
                        call error("WrongPSF","Wrong PSF number") 
                    end select      
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!               
                  else               
                    bg_image(ipoint,ix,iy) = bg_image(ipoint,ix,iy) + &
                       scts(ipoint,n)*psf4(energy,j,x1,y1, &
                       xms(n)-xoff(expip(i)),yms(n)-yoff(expip(i)), &
                       xcen(i),ycen(i),sigma(n),rcal) * &
                       exp_image(ipoint,ix,iy)
                  end if
                end if
               end if
              end do
            end do
            rate(ipoint,n)=scts(ipoint,n)
            erate(ipoint,n)=escts(ipoint,n)
            scts(ipoint,n)=scts(ipoint,n)*exp/expn     
            escts(ipoint,n)=escts(ipoint,n)*exp/expn 
          end do 
         end do
      end do
                   

!      write(*,*) "count_model:", count_model
                                     
      end subroutine mulsou   

  !----------------------------------------------------------------------
  subroutine rawcoords(ra,dec,nsrc,ra_pnt,dec_pnt,pa_pnt,datetime, &
                       chipix,chipiy,ccdinr)
  !----------------------------------------------------------------------

    use cal

    real(kind=double), dimension(max_fit), intent(in) :: ra, dec
    real(kind=double), intent(in) :: ra_pnt,dec_pnt,pa_pnt
    character(len=20), intent(in) :: datetime
    integer, intent(in) :: nsrc 
    integer, dimension(3,max_fit), intent(out) :: & 
             chipix, chipiy, ccdinr
    
    real(kind=double), dimension(:), pointer :: theta, phi, &
!!!    real(kind=single), dimension(:), pointer :: theta, phi, &
                       cc2x, cc2y, cc2z, cc1x, cc1y, cc1z      
    real(kind=double) :: timestamp
    integer(int16), dimension(:), pointer :: chipX, chipY
    integer(int8), dimension(:), pointer :: ccdnr  
    integer :: i, j, j0

    type(SpacecraftAttitudeType) :: myattitude  
    type (BoreSightT) :: boresight
    type(EquatorialDirectionType), allocatable :: srcDirs(:)

    allocate(srcdirs(nsrc))

    do i=1,nsrc
      srcDirs(i)%ra = ra(i)*pi/180d0
      srcDirs(i)%dec = dec(i)*pi/180d0
    end do

    timeStamp = toTimeTag(datetime) ! in caloalutils/caloaldefs.f90        
    boresight = CAL_getBoresight(TimeStamp)
    myattitude%ra=ra_pnt*pi/180d0
    myattitude%dec=dec_pnt*pi/180d0
    myattitude%apos=pa_pnt*pi/180d0
    do j=1,num_instr
      j0=expip(j) 
      if (j0==1) call CAL_setState(EPN)
      if (j0==2) call CAL_setState(EMOS1)
      if (j0==3) call CAL_setState(EMOS2) 
      call CAL_withSacCoordOrientation(boresight, myAttitude)
      call CAL_projectOntoTelCoord(boresight, srcDirs, theta, phi)
      call CAL_telCoordToCamCoord2(theta, phi, cc2x, cc2y, cc2z)
      call CAL_camCoord2ToCamCoord1(cc2x, cc2y, cc2z, cc1x, cc1y, cc1z)
      call CAL_camCoord1ToChipCoord(cc1x, cc1y, cc1z, chipX, chipY, ccdNr)
      do i=1,nsrc
        chipix(j0,i)=chipX(i-1)
        chipiy(j0,i)=chipY(i-1)
        ccdinr(j0,i)=ccdNr(i-1)
      end do
    end do
    deallocate(srcdirs)
   
  end subroutine rawcoords



  !----------------------------------------------------------------------
  subroutine ordr(nterms,nfix,iordr)
  !----------------------------------------------------------------------

    integer     i,iordr(max_par),io(max_par),nfix,nnfix,nterms

    if(iordr(1).eq.0)then
      do i=1,max_par
        iordr(i)=i
      end do
    endif

    do i=1,max_par
      if(iordr(i).eq.nfix) exit
      io(i)=iordr(i)
    end do

    nnfix=i

    do i=nnfix,nterms-1
      io(i)=iordr(i+1)
    end do

    io(nterms)=nfix

    if(nterms.ne.max_par) then
      do i=nterms+1,max_par
        io(i)=iordr(i)
      end do
    end if 

    iordr(1:max_par)=io(1:max_par)

  end subroutine ordr

  !----------------------------------------------------------------------
  subroutine pordr(nterms,a,b,iordr,mode)
  !----------------------------------------------------------------------

    real          a(max_par),b(max_par)
    integer       i,mode,nterms,iordr(max_par)

    if(mode.eq.1) then
      do i=1,nterms
        b(i)=a(iordr(i))
      enddo
    else

! andere richtung
      do i=1,nterms
        a(iordr(i))=b(i)
      enddo
    end if

  end subroutine pordr

  !----------------------------------------------------------------------
  function cage(array,n,vallo,valhi)
  !----------------------------------------------------------------------

    real ::        array(max_par),cage,value,vallo,valhi
    integer ::     ierr,n 

    ierr=0
!
! used to cage fit parameters to reasonable values
    value=array(n)

    if (value.lt.vallo) then
      array(n)=vallo
!      write(6,*)' allowed values ',vallo,valhi
    else if (value.gt.valhi) then
      array(n)=valhi
!      write(6,*)' allowed values ',vallo,valhi
    endif
    cage=array(n)

  end function cage    
      
  !----------------------------------------------------------------------
  subroutine cagexy (x1,y1,x0,y0,rcut2)
  !----------------------------------------------------------------------

    real     x0,y0,x1,y1,rcut2  
    real     ds

    ds=sqrt(((x1-x0)**2+(y1-y0)**2)/rcut2)
    if (ds.gt.1.) then
      x1 = x0+(x1-x0)/ds
      y1 = y0+(y1-y0)/ds
    end if
 
  end subroutine cagexy

  !----------------------------------------------------------------------
  subroutine like(lik,lik_arr,yfit1,yfit2)
  !----------------------------------------------------------------------

    real, dimension(max_events), intent(in) :: yfit1
    real, dimension(max_ima), intent(in) :: yfit2

    real, dimension(max_ima), intent(out) :: lik_arr
    real, intent(out) :: lik

    integer :: k,ipoint 

! accumulate statistics
    lik_arr(1:nim)=0.
    !loop over image pixels
    do k=1,npts
      ipoint=event(k)%ipt
      lik_arr(ipoint)=lik_arr(ipoint)+event(k)%n*log(max(yfit1(k),1e-30))
!!! try to prevent sources running into negative fluxes: 
     if (yfit1(k) < 0.0) then
         lik_arr(ipoint)=lik_arr(ipoint)-1.
    endif
!!!!!!! 
    end do

!!!!!!!    lik_arr(1:nim)=lik_arr(1:nim)-yfit2(1:nim)


    lik_arr(1:nim)=lik_arr(1:nim)-max(yfit2(1:nim),0.0)

!    lik_arr(1:nim)=-2.*lik_arr(1:nim)
!    remove factor 2 from ML => ML=-ln(P)
     lik_arr(1:nim)=-lik_arr(1:nim)

    lik=0. 
    do ipoint=1,nim
      lik=lik+lik_arr(ipoint)
    end do 

!   write(*,*) "like:",lik

   end subroutine like

  !----------------------------------------------------------------------
  subroutine mlfit(nterms,a,deltaa,sigmaa,iordr,xms,yms,ecut2,flamda, &
                   yfit1,like0,lik_arr,is,nfel,ierr)
  !----------------------------------------------------------------------

    integer :: i,j,k,minmethod
    integer :: ntermf,nterms,nfree,nfel,is,ierr,iordr(max_par)
    real(kind=double) :: det
    real :: a(max_par),deltaa(max_par),sigmaa(max_par)
    real :: xms(max_fit),yms(max_fit),b(max_par)
    real :: lik_arr(max_ima),yfit1(max_events),ytemp1(max_events)
    real :: yfit2(max_ima),yfit3,ytemp3
    !! real :: derv1(max_par,max_events),derv3(max_par)
    real :: derv3(max_par),testsig
    real, allocatable :: derv1(:,:)
    real :: like0,like1,like2,like3,like12,aj,ak,delta,delta2,deltak,ecut2,flamda
    real(kind=double) ::array(max_par,max_par),alpha(max_par,max_par), &
                        beta(max_par)      

    real, parameter :: small = 1e-6
    real :: test1,test2

!   small=tiny(small)  

    !allocate(derv1(max_par,max_events))
    allocate(derv1(nterms,npts))

    ntermf=nterms+nfel
    nfree=npts-nterms
    if (nfree.le.0) &
      call error("FitFailed","degrees of freedom less or equal 0")




!!!!!!!!!! insert 9/12/2003 !!!
    minmethod = 2


    if (minmethod .eq. 1) then 

       testsig=1.0

       beta(1:nterms)=0.
       alpha(1:nterms,1:nterms)=0. 

!       write(*,*) "b)"
       call funct(a,iordr,yfit1,yfit2,yfit3,is,nterms+nfel,xms,yms,ecut2)
       call like(like0,lik_arr,yfit1,yfit2)        
       do j=1,nterms
         aj=a(j)
         delta=deltaa(j)
         delta2=2.*deltaa(j) 
         a(j)=aj+delta 
!         write(*,*) "c)"
         call funct(a,iordr,yfit1,yfit2,yfit3,is,nterms+nfel,xms,yms,ecut2)
         call like(like1,lik_arr,yfit1,yfit2)
         a(j)=aj-delta 
!     write(*,*) "d)"       
         call funct(a,iordr,yfit1,yfit2,yfit3,is,nterms+nfel,xms,yms,ecut2)
         call like(like2,lik_arr,yfit1,yfit2)
         like12=like1-like2
         a(j)=aj      
         beta(j)=testsig*(-1.)*like12/delta2 
         alpha(j,j)=testsig*(like12-2.*like1+like0)/(delta*delta)
   ! derivatives del^2(like)/(del_ak del_aj) = ( like(aj + daj,ak+dak) - like(aj + daj,ak) - like(aj, ak+dak) + like(aj,ak) )/(daj*daj)          
         do k=1,j-j 
            ak=a(k)
            deltak=5*deltaa(k)
            a(k)=ak+deltak
            a(j)=aj+delta
! write(*,*) "e)"
            call funct(a,iordr,yfit1,yfit2,yfit3,is,nterms+nfel,xms,yms,ecut2)
            call like(like2,lik_arr,yfit1,yfit2)
            a(j)=aj
! write(*,*) "f)"
            call funct(a,iordr,yfit1,yfit2,yfit3,is,nterms+nfel,xms,yms,ecut2)
            call like(like3,lik_arr,yfit1,yfit2)
            alpha(j,k)=testsig*(like2-like1-like3+like0)/(delta*deltak)
            a(k)=ak
         end do
         do  k=1,j-1
           alpha(k,j)=alpha(j,k)
         end do
       end do
     else 
     !minmethod > 1

!!!!!!!!!! end insert 9/12/2003

!!!!!!! remove 9/12/2003 and replace with numerical alpha/ beta matrices"

    do j=1,nterms
      aj=a(j)

      delta=deltaa(j)
      delta2=2.0*delta

      a(j)=aj+delta
! write(*,*) "g)"
      call funct(a,iordr,yfit1,yfit2,yfit3,is,nterms+nfel,xms,yms,ecut2)

      a(j)=aj-delta
! write(*,*) "h)"
      call funct(a,iordr,ytemp1,yfit2,ytemp3,is,nterms+nfel,xms,yms,ecut2)

      derv1(j,1:npts)=(yfit1(1:npts)-ytemp1(1:npts))/delta2
      derv3(j)=(yfit3-ytemp3)/delta2
!      write(*,*) "j,derv1,derv3,delta2",j,derv1(j,1:npts),derv3(j),delta2
!      write(*,*) "a(1:7)",a(1:7) 

      a(j)=aj
    end do
!    
    beta(1:nterms)=0.
    alpha(1:nterms,1:nterms)=0.          
!
! write(*,*) "i)"
    call funct(a,iordr,yfit1,yfit2,yfit3,is,nterms+nfel,xms,yms,ecut2)
!
!      !      do i=1,npts
!      !       do j=1,nterms 
!      !          beta(j)=beta(j)+(event(i)%n/yfit1(i)*derv1(j,i)  &
!      !                                                     -derv3(j))
!      !          do k=1,j 
!      !            alpha(j,k)=alpha(j,k)+ &
!      !                   event(i)%n/yfit1(i)**2*derv1(j,i)*derv1(k,i)
!      !          end do
!      !        end do
!      !      end do
!
     do i=1,npts
       do j=1,nterms
!      !        beta(j)=beta(j)+(event(i)%n-yfit1(i))/max(event(i)%n,1)*derv1(j,i) 
        if (yfit1(i) .ne.  0.0) beta(j)=beta(j)+(event(i)%n-yfit1(i))/yfit1(i)*derv1(j,i)
        do k=1,j 
!      !         alpha(j,k)=alpha(j,k)+derv1(j,i)*derv1(k,i)/max(event(i)%n,1)
      
          if (yfit1(i) .ne.  0.0) alpha(j,k)=alpha(j,k)+derv1(j,i)*derv1(k,i)/yfit1(i)
        end do
      end do
    end do

    do j=1,nterms
      do k=1,j
        alpha(k,j)=alpha(j,k)
      end do
    end do
  endif
!!!! 9/12/2003
! write(*,*) "j)"
    call funct(a,iordr,yfit1,yfit2,yfit3,is,nterms+nfel,xms,yms,ecut2)
    call like(like1,lik_arr,yfit1,yfit2)

    do
      
      do j=1,nterms
        do k=1,nterms
          if(j.eq.k)then
            array(j,k)=1.0d0
          else 
            array(j,k)=alpha(j,k)/dsqrt(dabs(alpha(j,j))+small) &
                                 /dsqrt(dabs(alpha(k,k))+small)
          endif
        end do
        array(j,j)=1.0d0+flamda
      end do
  
      call mati(array,nterms,det)

!      b(1:nterms)=a(1:nterms)
      b = a
      
      do j=1,nterms
        do k=1,nterms
          b(j)=b(j)+beta(k)*array(j,k)/dsqrt(dabs(alpha(j,j))+small) &
                                      /dsqrt(dabs(alpha(k,k))+small)
!        write(*,*) "matrix:",j,k,beta(k),array(j,k),alpha(j,j),alpha(k,k)
        end do 
      end do

! write(*,*) "k)"
!!!! test output write(*,*):
!      if (.not.  ((b(1) >= 0.0) .or. (b(1) < 0.0))) then
!         write(*,*) "b:", b(1:10)
!         write(*,*) "beta:", beta(1:10)
!      do j=1,nterms
!        do k=1,nterms
!          write(*,*) "array:",j,k,array(j,k)
!          write(*,*) "alpha:",j,j,alpha(j,j)
!         write(*,*) "alpha:",k,k,alpha(k,k)
!          write(*,*) "sqrt:", j,k, dsqrt(dabs(alpha(j,j))+small) * dsqrt(dabs(alpha(k,k))+small)
!          write(*,*) "derv1:",j,k,derv1(j,k),yfit1(j)
!        write(*,*) "matrix:",j,k,beta(k),array(j,k),alpha(j,j),alpha(k,k)
!        end do 
!      end do
!      stop
!      endif 
      

!!!!!!!!!!!!!!!!!! end test output

      call funct(b,iordr,yfit1,yfit2,yfit3,is,nterms+nfel,xms,yms,ecut2)
      call like(like0,lik_arr,yfit1,yfit2)

      if (like0-like1.lt.0.) then
        exit
      else
        flamda=10.*flamda
        if (flamda.ge.1e10) exit
      end if

    end do

    a(1:nterms)=b(1:nterms)
    do j=1,nterms
      sigmaa(j)=dsqrt(dabs(array(j,j))/(dabs(alpha(j,j))+small))
    end do
    flamda=flamda/10.

    deallocate (derv1)


 


  end subroutine mlfit

  !----------------------------------------------------------------------
  subroutine funct(afit,iordr,yfit1,yfit2,yfit3,is,np,xms,yms,ecut2)
  !------------------------------------------------------------------------
      
    real :: yfit1(max_events),yfit2(max_ima),yfit3,a(max_par),afit(max_par), & 
            x0(max_fit),y0(max_fit),sig(max_fit),scts(max_ima,max_fit), &
            xms(max_fit),yms(max_fit),ecut2,ecut
    integer :: i,n,ierr,iordr(max_par),is,np, &
               ndat,npars

    ierr = 0

    call pordr(np,a,afit,iordr,-1)                               

    ecut = sqrt(ecut2)

! cage parameters
    npars=nim+3
!    write(*,*) '*****a:',a(1:14)
    do n=1,is 
      ndat=(n-1)*npars+1
      call cagexy (a(ndat),a(ndat+1),xms(n),yms(n),ecut2)
      x0(n)=a(ndat)
      y0(n)=a(ndat+1)
      ndat=(n-1)*npars+3
      sig(n)=cage(a,ndat,0.,maxext)
      do i=1,nim
        ndat=(n-1)*npars+i+3
!!!!!!!!!!!!!!!!!!!!!
        if (negfla) then 
          scts(i,n)=cage(a,ndat,-1.0,1e10)
        else
           scts(i,n)=cage(a,ndat,0.0,1e10) 
        endif
!         if  (scts(i,n) .eq. 0) then
!           write(*,*)  i,n, scts(i,n), x0(n),y0(n)
!         endif
!         scts(i,n)=cage(a,ndat,-1.,1e10)
!!!!!!!!!!!!!!!!!!!!!!
      end do
    end do 

! calculate expectation value of model + background

    call model(yfit1,yfit2,yfit3,x0,y0,sig,scts,is,ecut)

    call pordr(np,a,afit,iordr,1)                                

  end subroutine funct
           
  !----------------------------------------------------------------------
  subroutine model(yfit1,yfit2,yfit3,x0,y0,sig,scts,is,ecut)
  !----------------------------------------------------------------------
 
    real, dimension(max_events), intent(out) :: yfit1
    real, dimension(max_ima), intent(out) :: yfit2
    real, intent(out) :: yfit3
!    real, dimension(max_ima,max_fit,max_mask,max_mask), SAVE :: psf_save
    real, dimension(:,:,:,:), allocatable, SAVE :: psf_save


    real, dimension(max_fit), intent(in) :: x0, y0, sig
    real, dimension(max_ima,max_fit), intent(in) :: scts
!    logical, dimension(max_ima,max_fit) :: scts_ch 
!    logical, dimension(max_fit) ::  x0_ch, y0_ch, sig_ch

    real, dimension(max_ima,max_fit), SAVE :: scts_save
    real, dimension(max_ima,max_fit) :: scts_scale
    real, dimension(max_fit), SAVE  :: x0_save, y0_save, sig_save


    real, intent(in) :: ecut
    integer, intent(in) :: is
    
    real :: small
    real :: ene,x1,y1,ene_old
    real :: s1,s2,s3,fudge,stemp
    integer :: i,j,ii,jj,jpoint,l,n,ix,iy,ix0,ix1,iy0,iy1,ixx,iyy,ipoint,tempsum,xx0,yy0,ifound
    logical, dimension(max_ima,max_fit) :: scalepsf
    logical, save :: first = .true.


!    write(*,*) "model:",num_instr,num_image(1),num_image(2),x0(1),y0(1),sig(1),scts(1:4,1)
 

 
    if (first) allocate(psf_save(nim,is,max_mask,max_mask))
 

 
    small=1.E-10

    if (first) then
!      psf_save(1:max_ima,1:max_fit,1:max_mask,1:max_mask) = 0.0
      psf_save(1:nim,1:is,1:max_mask,1:max_mask) = 0.0
      scalepsf(1:max_ima,1:max_fit) = .false.
    else
!   check which parameters changed since last call: 
      do n=1,is
        if ((x0(n) .ne. x0_save(n)) .or. (y0(n) .ne. y0_save(n)) .or. (sig(n) .ne. sig_save(n))) then 
          scalepsf(1:max_ima,n) = .false.
          count_model=count_model+1
!          write(*,*) "parameters changed:", is,x0(n),x0_save(n),y0(n),y0_save(n),sig(n),sig_save(n)
        else
          do i=1,num_instr
            do j=1,num_image(i)
              ipoint=ip(i,j)
              if (scts(ipoint,n) .eq. 0.0) then
                scts_scale(ipoint,n) = 0.0
                scalepsf(ipoint,n) = .true.
              else if (scts(ipoint,n) .eq. scts_save(ipoint,n)) then
                scts_scale(ipoint,n) = 1.0
                scalepsf(ipoint,n) = .true.
              else if (abs(scts_save(ipoint,n)) .gt. small) then
                scts_scale(ipoint,n) =  scts(ipoint,n) / scts_save(ipoint,n)
!                write(*,*) ipoint,n,scts_scale(ipoint,n)
                scalepsf(ipoint,n) = .true.
              else
                scts_scale(ipoint,n) = 0.0
                scalepsf(ipoint,n) = .false.
                count_model=count_model+1
              endif
            enddo
          enddo
        endif
      enddo
    endif
!    write(*,*) "scalepsf:", scalepsf(1,1),scalepsf(1,2)

!!!! switch off scaling for testing purposes:
!          scalepsf(1:max_ima,1:max_fit) = .false.
!!!!!!
    
!    write(*,*) scalepsf(1:4,1),scts_scale(1:4,1) 
    

    l=1
    yfit3=0.
    fudge=1.0
    do i=1,num_instr


      if (num_evlist(i).eq.0) then

        s3=0.
        do j=1,num_image(i)
          ipoint=ip(i,j)
          ene=energy(ipoint)
!                    check whether same energy has been calculated for other instrument:
!                    CAL PSFs are identical for all EPIC inst. (as of 28/01/04),
!                    therefore it is pointless to call PSF again for same energy
!VVVVVVVVVVVVVVVVVV  remove, if separate PSF for EPIC instruments should be used
! removed, since fitting was slowed down (GL 03/03/04) 
!
!                    ifound=0
!                    do ii=1,i-1 
!                      do jj=1,num_image(ii)
!                         jpoint=ip(ii,jj)
!                         if (ene .eq. energy(jpoint)) then
!!                           write(*,*)  "found same energy:",ipoint,jpoint
!                           ifound=jpoint
!                           do n=1,is
!                             if  (.not. scalepsf(ipoint,n)) then
!                                
!!                               write(*,*)  "found same energy:",ipoint,jpoint
!                               scalepsf(ipoint,n) = .true.
!                               psf_save(ipoint,n,:,:)=psf_save(jpoint,n,:,:)
!                               scts_scale(ipoint,n) = 1.0
!                             endif
!                           end do
!                           exit
!                         endif
!                      end do
!                      if (ifound > 0) exit
!                   end do
!^^^^^^^^^^^^^^^^^^^
         
          s2=0.
          ix0 = mask%ix_min
          ix1 = mask%ix_max 
          iy0 = mask%iy_min
          iy1 = mask%iy_max

          tempsum=0
          yy0=0

          do iy = iy0, iy1
            yy0=yy0+1
            xx0=0
            do ix = ix0, ix1
              xx0=xx0+1
              if (mask%mask(ipoint,ix-ix0+1,iy-iy0+1).eq.1) then  
                tempsum=tempsum+1  
                x1=ix ! +0.5
                y1=iy ! +0.5
                s1=0.                  
                do n=1,is
                  if (scalepsf(ipoint,n) .and. (scts_scale(ipoint,n) .ne. 0.0)) then
                    s1=s1+psf_save(ipoint,n,xx0,yy0)*scts_scale(ipoint,n)
                    psf_save(ipoint,n,xx0,yy0)=psf_save(ipoint,n,xx0,yy0)*scts_scale(ipoint,n)
!                    write(*,*) "scaling PSF"
                    else
                    
!                    if (point_resp(i) .ge. 1 .and. point_resp(i) .le. 3) then
!                        stemp=scts(ipoint,n)*psf1(point_resp(i),energy,ipoint, &
!                              ix,iy,x0(n)-xoff(expip(i)),y0(n)-yoff(expip(i)), &
!                              xcen(i),ycen(i),sig(n), ecut)
!                        s1=s1+stemp
!                        psf_save(ipoint,n,xx0,yy0)=stemp
!                     else if (point_resp(i) .eq. 4) then
!                        stemp=scts(ipoint,n)*psf4(energy,ipoint, &
!                              x1,y1,x0(n)-xoff(expip(i)),y0(n)-yoff(expip(i)), &
!                              xcen(i),ycen(i),sig(n), ecut)
!                        s1=s1+stemp
!                       psf_save(ipoint,n,xx0,yy0)=stemp
!                     else
!                        call error("WrongPSF","Wrong PSF number") 
!                     endif
                   select case (point_resp(i))       
                     case(1)
                        stemp=scts(ipoint,n)*psf1(energy,ipoint, &
                              ix,iy,x0(n)-xoff(expip(i)),y0(n)-yoff(expip(i)), &
                             xcen(i),ycen(i),sig(n), ecut)
                        s1=s1+stemp
                        psf_save(ipoint,n,xx0,yy0)=stemp
                     case(2)
                        stemp=scts(ipoint,n)*psf2(energy,ipoint, &
                              ix,iy,x0(n)-xoff(expip(i)),y0(n)-yoff(expip(i)), &
                             xcen(i),ycen(i),sig(n), ecut)
                        s1=s1+stemp
                        psf_save(ipoint,n,xx0,yy0)=stemp
                     case(3)
                        stemp=scts(ipoint,n)*psf3(energy,ipoint, &
                              ix,iy,x0(n)-xoff(expip(i)),y0(n)-yoff(expip(i)), &
                             xcen(i),ycen(i),sig(n), ecut)
                        s1=s1+stemp
                        psf_save(ipoint,n,xx0,yy0)=stemp
                     case(4)
                        stemp=scts(ipoint,n)*psf4(energy,ipoint, &
                              x1,y1,x0(n)-xoff(expip(i)),y0(n)-yoff(expip(i)), &
                             xcen(i),ycen(i),sig(n), ecut)
                        s1=s1+stemp
                        psf_save(ipoint,n,xx0,yy0)=stemp
                     case default
                       call error("WrongPSF","Wrong PSF number") 
                  end select  
 
                  endif
                end do

                if (expfla) then
                  s1=s1*exp_image(ipoint,ix,iy)
                end if
                ixx=min(max(ix,1),npix(1))
                iyy=min(max(iy,1),npix(2))       
                s1=s1*fudge+bg_image(ipoint,ixx,iyy) 
                yfit1(l)=s1
                s2=s2+s1
                l=l+1
    
              end if

            end do
          end do


          yfit2(ipoint)=s2   
          s3=s3+s2

        end do

        yfit3=yfit3+s3
     

      else

      end if

    end do          

    first = .false.
    x0_save=x0
    y0_save=y0
    sig_save=sig
    scts_save=scts

  end subroutine model





!%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% powmin.f %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 
SUBROUTINE powmin(n, rhobeg, rhoend, iprint, maxfun, a, iordr, xms, yms, &
                  ecut2, yfit1, like0, lik_arr, is, nfel, ierr)

INTEGER, PARAMETER  :: dp = SELECTED_REAL_KIND(12, 60)
INTEGER, INTENT(IN)        :: n
!REAL (dp)                  :: x(n)
REAL (dp), INTENT(IN)      :: rhobeg
REAL (dp), INTENT(IN)      :: rhoend
INTEGER, INTENT(IN)        :: iprint
INTEGER, INTENT(IN)        :: maxfun
! additional parameters for emldetect use:
REAL, INTENT(IN)           :: ecut2
REAL                       :: a(max_par), xms(max_fit), yms(max_fit), lik_arr(max_ima)
integer                    :: iordr(max_par)
integer                    :: is, nfel, ierr 
real                       :: yfit1(max_events), like0

!     This subroutine seeks the least value of a function of many variables,
!     by a trust region method that forms quadratic models by interpolation.
!     The algorithm is described in "UOBYQA: unconstrained optimization by
!     quadratic approximation" by M.J.D. Powell, Report DAMTP 2000/NA14,
!     University of Cambridge. The arguments of the subroutine are as follows.

!     N must be set to the number of variables and must be at least two.
!     Initial values of the variables must be set in X(1),X(2),...,X(N). They
!       will be changed to the values that give the least calculated F.
!     RHOBEG and RHOEND must be set to the initial and final values of a trust
!       region radius, so both must be positive with RHOEND<=RHOBEG. Typically
!       RHOBEG should be about one tenth of the greatest expected change to a
!       variable, and RHOEND should indicate the accuracy that is required in
!       the final values of the variables.
!     The value of IPRINT should be set to 0, 1, 2 or 3, which controls the
!       amount of printing.  Specifically, there is no output if IPRINT=0 and
!       there is output only at the return if IPRINT=1.  Otherwise, each new
!       value of RHO is printed, with the best vector of variables so far and
!       the corresponding value of the objective function. Further, each new
!       value of F with its variables are output if IPRINT=3.
!     MAXFUN must be set to an upper bound on the number of calls of CALFUN.
!     The array W will be used for working space. Its length must be at least
!       ( N**4 + 8*N**3 + 23*N**2 + 42*N + max [ 2*N**2 + 4, 18*N ] ) / 4.

!     SUBROUTINE CALFUN (N,X,F) must be provided by the user. It must set F to
!     the value of the objective function for the variables X(1),X(2),...,X(N).

INTEGER  :: npt

!     Partition the working space array, so that different parts of it can be
!     treated separately by the subroutine that performs the main calculation.

npt = (n*n + 3*n + 2) / 2
CALL uobyqb(n, rhobeg, rhoend, iprint, maxfun, npt,  a, iordr, xms, yms, &
                  ecut2, yfit1, like0, lik_arr, is, nfel, ierr)
RETURN
END SUBROUTINE powmin

!%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% uobyqb.f %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

SUBROUTINE uobyqb(n, rhobeg, rhoend, iprint, maxfun, npt,  a, iordr, xms, yms, &
                  ecut2, yfit1, like0, lik_arr, is, nfel, ierr)

INTEGER, PARAMETER  :: dp = SELECTED_REAL_KIND(12, 60)
INTEGER, INTENT(IN)        :: n
REAL (dp)                  :: x(n)
REAL (dp), INTENT(IN)      :: rhobeg
REAL (dp), INTENT(IN)      :: rhoend
INTEGER, INTENT(IN)        :: iprint
INTEGER, INTENT(IN)        :: maxfun
INTEGER, INTENT(IN)        :: npt
! additional parameters for emldetect use:
REAL                       :: a(max_par), xms(max_fit), yms(max_fit), lik_arr(max_ima)
integer                    :: iordr(max_par)
integer                    :: is, nfel, ierr 
real                       :: yfit1(max_events), yfit2(max_ima),yfit3,like0,ecut2


! The following arrays were previously passed as arguments:

REAL (dp)  :: xbase(n), xopt(n), xnew(n), xpt(npt,n), pq(npt-1)
REAL (dp)  :: pl(npt,npt-1), h(n,n), g(n), d(n), vlag(npt), w(npt)

!     The arguments N, X, RHOBEG, RHOEND, IPRINT and MAXFUN are identical to
!       the corresponding arguments in SUBROUTINE UOBYQA.

!     NPT is set by UOBYQA to (N*N+3*N+2)/2 for the above dimension statement.
!     XBASE will contain a shift of origin that reduces the contributions from
!       rounding errors to values of the model and Lagrange functions.
!     XOPT will be set to the displacement from XBASE of the vector of
!       variables that provides the least calculated F so far.
!     XNEW will be set to the displacement from XBASE of the vector of
!       variables for the current calculation of F.
!     XPT will contain the interpolation point coordinates relative to XBASE.
!     PQ will contain the parameters of the quadratic model.
!     PL will contain the parameters of the Lagrange functions.
!     H will provide the second derivatives that TRSTEP and LAGMAX require.
!     G will provide the first derivatives that TRSTEP and LAGMAX require.
!     D is reserved for trial steps from XOPT, except that it will contain
!       diagonal second derivatives during the initialization procedure.
!     VLAG will contain the values of the Lagrange functions at a new point X.
!     The array W will be used for working space.

REAL (dp)  :: half = 0.5_dp, one = 1.0_dp, tol = 0.01_dp, two = 2.0_dp
REAL (dp)  :: zero = 0.0_dp
REAL (dp)  :: ddknew, delta, detrat, diff, distest, dnorm, errtol, estim
REAL (dp)  :: evalue, f, fbase, fopt, fsave, ratio, rho, rhosq, sixthm
REAL (dp)  :: sum, sumg, sumh, temp, tempa, tworsq, vmax, vquad, wmult
INTEGER    :: i, ih, ip, iq, iw, j, jswitch, k, knew, kopt, ksave, ktemp
INTEGER    :: nf, nftest, nnp, nptm

!     Set some constants.

nnp = n + n + 1
nptm = npt - 1
nftest = MAX(maxfun,1)

!     Initialization. NF is the number of function calculations so far.

rho = rhobeg
rhosq = rho * rho
nf = 0
DO  i = 1, n
  x(i)=dble(a(i))
  xbase(i) = x(i)
  xpt(1:npt,i) = zero
END DO
pl(1:npt,1:nptm) = zero

!     The branch to label 120 obtains a new value of the objective function
!     and then there is a branch back to label 50, because the new function
!     value is needed to form the initial quadratic model. The least function
!     value so far and its index are noted below.

50 x(1:n) = xbase(1:n) + xpt(nf+1,1:n)
GO TO 150

70 IF (nf == 1) THEN
  fopt = f
  kopt = nf
  fbase = f
  j = 0
  jswitch = -1
  ih = n
ELSE
  IF (f < fopt) THEN
    fopt = f
    kopt = nf
  END IF
END IF

!     Form the gradient and diagonal second derivatives of the initial
!     quadratic model and Lagrange functions.

IF (nf <= nnp) THEN
  jswitch = -jswitch
  IF (jswitch > 0) THEN
    IF (j >= 1) THEN
      ih = ih + j
      IF (w(j) < zero) THEN
        d(j) = (fsave+f-two*fbase) / rhosq
        pq(j) = (fsave-f) / (two*rho)
        pl(1,ih) = -two / rhosq
        pl(nf-1,j) = half / rho
        pl(nf-1,ih) = one / rhosq
      ELSE
        pq(j) = (4.0D0*fsave-3.0D0*fbase-f) / (two*rho)
        d(j) = (fbase+f-two*fsave) / rhosq
        pl(1,j) = -1.5D0 / rho
        pl(1,ih) = one / rhosq
        pl(nf-1,j) = two / rho
        pl(nf-1,ih) = -two / rhosq
      END IF
      pq(ih) = d(j)
      pl(nf,j) = -half / rho
      pl(nf,ih) = one / rhosq
    END IF
    
!     Pick the shift from XBASE to the next initial interpolation point
!     that provides diagonal second derivatives.
    
    IF (j < n) THEN
      j = j + 1
      xpt(nf+1,j) = rho
    END IF
  ELSE
    fsave = f
    IF (f < fbase) THEN
      w(j) = rho
      xpt(nf+1,j) = two * rho
    ELSE
      w(j) = -rho
      xpt(nf+1,j) = -rho
    END IF
  END IF
  IF (nf < nnp) GO TO 50
  
!     Form the off-diagonal second derivatives of the initial quadratic model.
  
  ih = n
  ip = 1
  iq = 2
END IF
ih = ih + 1
IF (nf > nnp) THEN
  temp = one / (w(ip)*w(iq))
  tempa = f - fbase - w(ip) * pq(ip) - w(iq) * pq(iq)
  pq(ih) = (tempa - half*rhosq*(d(ip)+d(iq))) * temp
  pl(1,ih) = temp
  iw = ip + ip
  IF (w(ip) < zero) iw = iw + 1
  pl(iw,ih) = -temp
  iw = iq + iq
  IF (w(iq) < zero) iw = iw + 1
  pl(iw,ih) = -temp
  pl(nf,ih) = temp
  
!     Pick the shift from XBASE to the next initial interpolation point
!     that provides off-diagonal second derivatives.
  
  ip = ip + 1
END IF
IF (ip == iq) THEN
  ih = ih + 1
  ip = 1
  iq = iq + 1
END IF
IF (nf < npt) THEN
  xpt(nf+1,ip) = w(ip)
  xpt(nf+1,iq) = w(iq)
  GO TO 50
END IF

!     Set parameters to begin the iterations for the current RHO.

sixthm = zero
delta = rho
80 tworsq = (two*rho) ** 2
rhosq = rho * rho

!     Form the gradient of the quadratic model at the trust region centre.

90 knew = 0
ih = n
DO  j = 1, n
  xopt(j) = xpt(kopt,j)
  g(j) = pq(j)
  DO  i = 1, j
    ih = ih + 1
    g(i) = g(i) + pq(ih) * xopt(j)
    IF (i < j) g(j) = g(j) + pq(ih) * xopt(i)
    h(i,j) = pq(ih)
  END DO
END DO

!     Generate the next trust region step and test its length.  Set KNEW
!     to -1 if the purpose of the next F will be to improve conditioning,
!     and also calculate a lower bound on the Hessian term of the model Q.

CALL trstep(n, g, h, delta, tol, d, evalue)
temp = zero
DO i = 1, n
  temp = temp + d(i)**2
END DO
dnorm = MIN(delta,SQRT(temp))
errtol = -one
IF (dnorm < half*rho) THEN
  knew = -1
  errtol = half * evalue * rho * rho
  IF (nf <= npt+9) errtol = zero
  GO TO 290
END IF

!     Calculate the next value of the objective function.

130 DO  i = 1, n
  xnew(i) = xopt(i) + d(i)
  x(i) = xbase(i) + xnew(i)
END DO
150 IF (nf >= nftest) THEN
  IF (iprint > 0) WRITE(*, 5000)
  GO TO 420
END IF
nf = nf + 1

a(1:n)=real(x(1:n))

!CALL calfun(n, x, f)
call calfun (n, a, f, iordr,yfit1,yfit2,yfit3,is,xms,yms,ecut2,lik_arr,nfel)
!write(*,*) "calfun:", f
like0=f
IF (iprint == 3) THEN
  WRITE(*, 5100) nf, f, x(1:n)
END IF
IF (nf <= npt) GO TO 70
IF (knew == -1) GO TO 420

!     Use the quadratic model to predict the change in F due to the step D,
!     and find the values of the Lagrange functions at the new point.

vquad = zero
ih = n
DO  j = 1, n
  w(j) = d(j)
  vquad = vquad + w(j) * pq(j)
  DO  i = 1, j
    ih = ih + 1
    w(ih) = d(i) * xnew(j) + d(j) * xopt(i)
    IF (i == j) w(ih) = half * w(ih)
    vquad = vquad + w(ih) * pq(ih)
  END DO
END DO
DO  k = 1, npt
  temp = zero
  DO  j = 1, nptm
    temp = temp + w(j) * pl(k,j)
  END DO
  vlag(k) = temp
END DO
vlag(kopt) = vlag(kopt) + one

!     Update SIXTHM, which is a lower bound on one sixth of the greatest
!     third derivative of F.

diff = f - fopt - vquad
sum = zero
DO  k = 1, npt
  temp = zero
  DO  i = 1, n
    temp = temp + (xpt(k,i)-xnew(i)) ** 2
  END DO
  temp = SQRT(temp)
  sum = sum + ABS(temp*temp*temp*vlag(k))
END DO
sixthm = MAX(sixthm, ABS(diff)/sum)

!     Update FOPT and XOPT if the new F is the least value of the objective
!     function so far.  Then branch if D is not a trust region step.

fsave = fopt
IF (f < fopt) THEN
  fopt = f
  xopt(1:n) = xnew(1:n)
END IF
ksave = knew
IF (knew <= 0) THEN
  
!     Pick the next value of DELTA after a trust region step.
  
  IF (vquad >= zero) THEN
    IF (iprint > 0) WRITE(*, 5200)
    GO TO 420
  END IF
  ratio = (f-fsave) / vquad
  IF (ratio <= 0.1D0) THEN
    delta = half * dnorm
  ELSE IF (ratio <= 0.7D0) THEN
    delta = MAX(half*delta,dnorm)
  ELSE
    delta = MAX(delta, 1.25D0*dnorm, dnorm+rho)
  END IF
  IF (delta <= 1.5D0*rho) delta = rho
  
!     Set KNEW to the index of the next interpolation point to be deleted.
  
  ktemp = 0
  detrat = zero
  IF (f >= fsave) THEN
    ktemp = kopt
    detrat = one
  END IF
  DO  k = 1, npt
    sum = zero
    DO  i = 1, n
      sum = sum + (xpt(k,i)-xopt(i)) ** 2
    END DO
    temp = ABS(vlag(k))
    IF (sum > rhosq) temp = temp * (sum/rhosq) ** 1.5D0
    IF (temp > detrat .AND. k /= ktemp) THEN
      detrat = temp
      ddknew = sum
      knew = k
    END IF
  END DO
  IF (knew == 0) GO TO 290
END IF

!     Replace the interpolation point that has index KNEW by the point XNEW,
!     and also update the Lagrange functions and the quadratic model.

DO  i = 1, n
  xpt(knew,i) = xnew(i)
END DO
temp = one / vlag(knew)
DO  j = 1, nptm
  pl(knew,j) = temp * pl(knew,j)
  pq(j) = pq(j) + diff * pl(knew,j)
END DO
DO  k = 1, npt
  IF (k /= knew) THEN
    temp = vlag(k)
    DO  j = 1, nptm
      pl(k,j) = pl(k,j) - temp * pl(knew,j)
    END DO
  END IF
END DO

!     Update KOPT if F is the least calculated value of the objective function.
!     Then branch for another trust region calculation.  The case KSAVE > 0
!     indicates that a model step has just been taken.

IF (f < fsave) THEN
  kopt = knew
  GO TO 90
END IF
IF (ksave > 0) GO TO 90
IF (dnorm > two*rho) GO TO 90
IF (ddknew > tworsq) GO TO 90

!     Alternatively, find out if the interpolation points are close
!     enough to the best point so far.

290 DO  k = 1, npt
  w(k) = zero
  DO  i = 1, n
    w(k) = w(k) + (xpt(k,i)-xopt(i)) ** 2
  END DO
END DO
320 knew = -1
distest = tworsq
DO  k = 1, npt
  IF (w(k) > distest) THEN
    knew = k
    distest = w(k)
  END IF
END DO

!     If a point is sufficiently far away, then set the gradient and Hessian
!     of its Lagrange function at the centre of the trust region, and find
!     half the sum of squares of components of the Hessian.

IF (knew > 0) THEN
  ih = n
  sumh = zero
  DO  j = 1, n
    g(j) = pl(knew,j)
    DO  i = 1, j
      ih = ih + 1
      temp = pl(knew,ih)
      g(j) = g(j) + temp * xopt(i)
      IF (i < j) THEN
        g(i) = g(i) + temp * xopt(j)
        sumh = sumh + temp * temp
      END IF
      h(i,j) = temp
    END DO
    sumh = sumh + half * temp * temp
  END DO
  
!     If ERRTOL is positive, test whether to replace the interpolation point
!     with index KNEW, using a bound on the maximum modulus of its Lagrange
!     function in the trust region.
  
  IF (errtol > zero) THEN
    w(knew) = zero
    sumg = zero
    DO  i = 1, n
      sumg = sumg + g(i) ** 2
    END DO
    estim = rho * (SQRT(sumg)+rho*SQRT(half*sumh))
    wmult = sixthm * distest ** 1.5D0
    IF (wmult*estim <= errtol) GO TO 320
  END IF
  
!     If the KNEW-th point may be replaced, then pick a D that gives a large
!     value of the modulus of its Lagrange function within the trust region.
!     Here the vector XNEW is used as temporary working space.
  
  CALL lagmax(n, g, h, rho, d, xnew, vmax)
  IF (errtol > zero) THEN
    IF (wmult*vmax <= errtol) GO TO 320
  END IF
  GO TO 130
END IF
IF (dnorm > rho) GO TO 90

!     Prepare to reduce RHO by shifting XBASE to the best point so far,
!     and make the corresponding changes to the gradients of the Lagrange
!     functions and the quadratic model.

IF (rho > rhoend) THEN
  ih = n
  DO  j = 1, n
    xbase(j) = xbase(j) + xopt(j)
    DO  k = 1, npt
      xpt(k,j) = xpt(k,j) - xopt(j)
    END DO
    DO  i = 1, j
      ih = ih + 1
      pq(i) = pq(i) + pq(ih) * xopt(j)
      IF (i < j) THEN
        pq(j) = pq(j) + pq(ih) * xopt(i)
        DO  k = 1, npt
          pl(k,j) = pl(k,j) + pl(k,ih) * xopt(i)
        END DO
      END IF
      DO  k = 1, npt
        pl(k,i) = pl(k,i) + pl(k,ih) * xopt(j)
      END DO
    END DO
  END DO
  
!     Pick the next values of RHO and DELTA.
  
  delta = half * rho
  ratio = rho / rhoend
  IF (ratio <= 16.0D0) THEN
    rho = rhoend
  ELSE IF (ratio <= 250.0D0) THEN
    rho = SQRT(ratio) * rhoend
  ELSE
    rho = 0.1D0 * rho
  END IF
  delta = MAX(delta,rho)
  IF (iprint >= 2) THEN
    IF (iprint >= 3) WRITE(*, 5300)
    WRITE(*, 5400) rho, nf
    WRITE(*, 5500) fopt, xbase(1:n)
  END IF
  GO TO 80
END IF

!     Return from the calculation, after another Newton-Raphson step, if
!     it is too short to have been tried before.

IF (errtol >= zero) GO TO 130
420 IF (fopt <= f) THEN
  DO  i = 1, n
    x(i) = xbase(i) + xopt(i)
  END DO
  f = fopt
END IF
IF (iprint >= 1) THEN
  WRITE(*, 5600) nf
  WRITE(*, 5500) f, x(1:n)
END IF
RETURN

5000 FORMAT (/T5, 'Return from UOBYQA because CALFUN has been',  &
    ' called MAXFUN times')
5100 FORMAT (/T5, 'Function number',i6,'    F =', g18.10,  &
    '    The corresponding X is:'/ (t3, 5g15.6))
5200 FORMAT (/T5, 'Return from UOBYQA because a trust',  &
    ' region step has failed to reduce Q')
5300 FORMAT (' ')
5400 FORMAT (/T5, 'New RHO =', g11.4, '     Number of function values =',i6)
5500 FORMAT (T5, 'Least value of F =', g23.15,  &
    '         The corresponding X is:'/ (t3, 5g15.6))
5600 FORMAT (/T5, 'At the return from UOBYQA',  &
    '     Number of function values =', i6)
END SUBROUTINE uobyqb

!%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% trstep.f %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

SUBROUTINE trstep(n, g, h, delta, tol, d, evalue)


INTEGER, PARAMETER  :: dp = SELECTED_REAL_KIND(12, 60)
INTEGER, INTENT(IN)        :: n
REAL (dp), INTENT(IN)      :: g(:)
REAL (dp), INTENT(IN OUT)  :: h(:,:)
REAL (dp), INTENT(IN)      :: delta
REAL (dp), INTENT(IN)      :: tol
REAL (dp), INTENT(OUT)     :: d(:)
REAL (dp), INTENT(OUT)     :: evalue

!     N is the number of variables of a quadratic objective function, Q say.
!     G is the gradient of Q at the origin.
!     H is the Hessian matrix of Q.  Only the upper triangular and diagonal
!       parts need be set.  The lower triangular part is used to store the
!       elements of a Householder similarity transformation.
!     DELTA is the trust region radius, and has to be positive.
!     TOL is the value of a tolerance from the open interval (0,1).
!     D will be set to the calculated vector of variables.

!     EVALUE will be set to the least eigenvalue of H if and only if D is a
!     Newton-Raphson step.  Then EVALUE will be positive, but otherwise it
!     will be set to zero.

!     Let MAXRED be the maximum of Q(0)-Q(D) subject to ||D|| <= DELTA,
!     and let ACTRED be the value of Q(0)-Q(D) that is actually calculated.
!     We take the view that any D is acceptable if it has the properties

!             ||D|| <= DELTA  and  ACTRED <= (1-TOL)*MAXRED.

!     The calculation of D is done by the method of Section 2 of the paper
!     by MJDP in the 1997 Dundee Numerical Analysis Conference Proceedings,
!     after transforming H to tridiagonal form.

!     The arrays GG, TD, TN, W, PIV and Z will be used for working space.
REAL (dp)  :: gg(n), td(n), tn(n), w(n), piv(n), z(n)

REAL (dp)  :: delsq, dhd, dnorm, dsq, dtg, dtz, gam, gnorm, gsq, hnorm
REAL (dp)  :: par, parl, parlest, paru, paruest, phi, phil, phiu, pivksv
REAL (dp)  :: pivot, posdef, scale, shfmax, shfmin, shift, slope, sum
REAL (dp)  :: tdmin, temp, tempa, tempb, wsq, wwsq, wz, zsq
INTEGER    :: i, iterc, j, jp, k, kp, kpp, ksav, ksave, nm
REAL (dp)  :: one = 1.0_dp, two = 2.0_dp, zero = 0.0_dp

!     Initialization.

delsq = delta * delta
evalue = zero
nm = n - 1
DO  i = 1, n
  d(i) = zero
  td(i) = h(i,i)
  DO  j = 1, i
    h(i,j) = h(j,i)
  END DO
END DO

!     Apply Householder transformations to obtain a tridiagonal matrix that
!     is similar to H, and put the elements of the Householder vectors in
!     the lower triangular part of H.  Further, TD and TN will contain the
!     diagonal and other nonzero elements of the tridiagonal matrix.

DO  k = 1, nm
  kp = k + 1
  sum = zero
  IF (kp < n) THEN
    kpp = kp + 1
    DO  i = kpp, n
      sum = sum + h(i,k) ** 2
    END DO
  END IF
  IF (sum == zero) THEN
    tn(k) = h(kp,k)
    h(kp,k) = zero
  ELSE
    temp = h(kp,k)
    tn(k) = SIGN(SQRT(sum+temp*temp),temp)
    h(kp,k) = -sum / (temp+tn(k))
    temp = SQRT(two/(sum+h(kp,k)**2))
    DO  i = kp, n
      w(i) = temp * h(i,k)
      h(i,k) = w(i)
      z(i) = td(i) * w(i)
    END DO
    wz = zero
    DO  j = kp, nm
      jp = j + 1
      DO  i = jp, n
        z(i) = z(i) + h(i,j) * w(j)
        z(j) = z(j) + h(i,j) * w(i)
      END DO
      wz = wz + w(j) * z(j)
    END DO
    wz = wz + w(n) * z(n)
    DO  j = kp, n
      td(j) = td(j) + w(j) * (wz*w(j)-two*z(j))
      IF (j < n) THEN
        jp = j + 1
        DO  i = jp, n
          h(i,j) = h(i,j) - w(i) * z(j) - w(j) * (z(i)-wz*w(i))
        END DO
      END IF
    END DO
  END IF
END DO

!     Form GG by applying the similarity transformation to G.

gsq = zero
DO  i = 1, n
  gg(i) = g(i)
  gsq = gsq + g(i) ** 2
END DO
gnorm = SQRT(gsq)
DO  k = 1, nm
  kp = k + 1
  sum = zero
  DO  i = kp, n
    sum = sum + gg(i) * h(i,k)
  END DO
  DO  i = kp, n
    gg(i) = gg(i) - sum * h(i,k)
  END DO
END DO

!     Begin the trust region calculation with a tridiagonal matrix by
!     calculating the norm of H. Then treat the case when H is zero.

hnorm = ABS(td(1)) + ABS(tn(1))
tdmin = td(1)
tn(n) = zero
DO  i = 2, n
  temp = ABS(tn(i-1)) + ABS(td(i)) + ABS(tn(i))
  hnorm = MAX(hnorm,temp)
  tdmin = MIN(tdmin,td(i))
END DO
IF (hnorm == zero) THEN
  IF (gnorm == zero) GO TO 420
  scale = delta / gnorm
  DO  i = 1, n
    d(i) = -scale * gg(i)
  END DO
  GO TO 380
END IF

!     Set the initial values of PAR and its bounds.

parl = MAX(zero, -tdmin, gnorm/delta-hnorm)
parlest = parl
par = parl
paru = zero
paruest = zero
posdef = zero
iterc = 0

!     Calculate the pivots of the Cholesky factorization of (H+PAR*I).

160 iterc = iterc + 1
ksav = 0
piv(1) = td(1) + par
k = 1
170 IF (piv(k) > zero) THEN
  piv(k+1) = td(k+1) + par - tn(k) ** 2 / piv(k)
ELSE
  IF (piv(k) < zero .OR. tn(k) /= zero) GO TO 180
  ksav = k
  piv(k+1) = td(k+1) + par
END IF
k = k + 1
IF (k < n) GO TO 170
IF (piv(k) >= zero) THEN
  IF (piv(k) == zero) ksav = k
  
!     Branch if all the pivots are positive, allowing for the case when
!     G is zero.
  
  IF (ksav == 0 .AND. gsq > zero) GO TO 250
  IF (gsq == zero) THEN
    IF (par == zero) GO TO 380
    paru = par
    paruest = par
    IF (ksav == 0) GO TO 210
  END IF
  k = ksav
END IF

!     Set D to a direction of nonpositive curvature of the given tridiagonal
!     matrix, and thus revise PARLEST.

180 d(k) = one
IF (ABS(tn(k)) <= ABS(piv(k))) THEN
  dsq = one
  dhd = piv(k)
ELSE
  temp = td(k+1) + par
  IF (temp <= ABS(piv(k))) THEN
    d(k+1) = SIGN(one,-tn(k))
    dhd = piv(k) + temp - two * ABS(tn(k))
  ELSE
    d(k+1) = -tn(k) / temp
    dhd = piv(k) + tn(k) * d(k+1)
  END IF
  dsq = one + d(k+1) ** 2
END IF
190 IF (k > 1) THEN
  k = k - 1
  IF (tn(k) /= zero) THEN
    d(k) = -tn(k) * d(k+1) / piv(k)
    dsq = dsq + d(k) ** 2
    GO TO 190
  END IF
  d(1:k) = zero
END IF
parl = par
parlest = par - dhd / dsq

!     Terminate with D set to a multiple of the current D if the following
!     test suggests that it suitable to do so.

210 temp = paruest
IF (gsq == zero) temp = temp * (one-tol)
IF (paruest > zero .AND. parlest >= temp) THEN
  dtg = DOT_PRODUCT( d(1:n), gg(1:n) )
  scale = -SIGN(delta/SQRT(dsq),dtg)
  d(1:n) = scale * d(1:n)
  GO TO 380
END IF

!     Pick the value of PAR for the next iteration.

240 IF (paru == zero) THEN
  par = two * parlest + gnorm / delta
ELSE
  par = 0.5D0 * (parl+paru)
  par = MAX(par,parlest)
END IF
IF (paruest > zero) par = MIN(par,paruest)
GO TO 160

!     Calculate D for the current PAR in the positive definite case.

250 w(1) = -gg(1) / piv(1)
DO  i = 2, n
  w(i) = (-gg(i)-tn(i-1)*w(i-1)) / piv(i)
END DO
d(n) = w(n)
DO  i = nm, 1, -1
  d(i) = w(i) - tn(i) * d(i+1) / piv(i)
END DO

!     Branch if a Newton-Raphson step is acceptable.

dsq = zero
wsq = zero
DO  i = 1, n
  dsq = dsq + d(i) ** 2
  wsq = wsq + piv(i) * w(i) ** 2
END DO
IF (par /= zero .OR. dsq > delsq) THEN
  
!     Make the usual test for acceptability of a full trust region step.
  
  dnorm = SQRT(dsq)
  phi = one / dnorm - one / delta
  temp = tol * (one+par*dsq/wsq) - dsq * phi * phi
  IF (temp >= zero) THEN
    scale = delta / dnorm
    DO  i = 1, n
      d(i) = scale * d(i)
    END DO
    GO TO 380
  END IF
  IF (iterc >= 2 .AND. par <= parl) GO TO 380
  IF (paru > zero .AND. par >= paru) GO TO 380
  
!     Complete the iteration when PHI is negative.
  
  IF (phi < zero) THEN
    parlest = par
    IF (posdef == one) THEN
      IF (phi <= phil) GO TO 380
      slope = (phi-phil) / (par-parl)
      parlest = par - phi / slope
    END IF
    slope = one / gnorm
    IF (paru > zero) slope = (phiu-phi) / (paru-par)
    temp = par - phi / slope
    IF (paruest > zero) temp = MIN(temp,paruest)
    paruest = temp
    posdef = one
    parl = par
    phil = phi
    GO TO 240
  END IF
  
!     If required, calculate Z for the alternative test for convergence.
  
  IF (posdef == zero) THEN
    w(1) = one / piv(1)
    DO  i = 2, n
      temp = -tn(i-1) * w(i-1)
      w(i) = (SIGN(one,temp)+temp) / piv(i)
    END DO
    z(n) = w(n)
    DO  i = nm, 1, -1
      z(i) = w(i) - tn(i) * z(i+1) / piv(i)
    END DO
    wwsq = zero
    zsq = zero
    dtz = zero
    DO  i = 1, n
      wwsq = wwsq + piv(i) * w(i) ** 2
      zsq = zsq + z(i) ** 2
      dtz = dtz + d(i) * z(i)
    END DO
    
!     Apply the alternative test for convergence.
    
    tempa = ABS(delsq-dsq)
    tempb = SQRT(dtz*dtz+tempa*zsq)
    gam = tempa / (SIGN(tempb,dtz)+dtz)
    temp = tol * (wsq+par*delsq) - gam * gam * wwsq
    IF (temp >= zero) THEN
      DO  i = 1, n
        d(i) = d(i) + gam * z(i)
      END DO
      GO TO 380
    END IF
    parlest = MAX(parlest,par-wwsq/zsq)
  END IF
  
!     Complete the iteration when PHI is positive.
  
  slope = one / gnorm
  IF (paru > zero) THEN
    IF (phi >= phiu) GO TO 380
    slope = (phiu-phi) / (paru-par)
  END IF
  parlest = MAX(parlest,par-phi/slope)
  paruest = par
  IF (posdef == one) THEN
    slope = (phi-phil) / (par-parl)
    paruest = par - phi / slope
  END IF
  paru = par
  phiu = phi
  GO TO 240
END IF

!     Set EVALUE to the least eigenvalue of the second derivative matrix if
!     D is a Newton-Raphson step. SHFMAX will be an upper bound on EVALUE.

shfmin = zero
pivot = td(1)
shfmax = pivot
DO  k = 2, n
  pivot = td(k) - tn(k-1) ** 2 / pivot
  shfmax = MIN(shfmax,pivot)
END DO

!     Find EVALUE by a bisection method, but occasionally SHFMAX may be
!     adjusted by the rule of false position.

ksave = 0
350 shift = 0.5D0 * (shfmin+shfmax)
k = 1
temp = td(1) - shift

360 IF (temp > zero) THEN
  piv(k) = temp
  IF (k < n) THEN
    temp = td(k+1) - shift - tn(k) ** 2 / temp
    k = k + 1
    GO TO 360
  END IF
  shfmin = shift
ELSE
  IF (k < ksave) GO TO 370
  IF (k == ksave) THEN
    IF (pivksv == zero) GO TO 370
    IF (piv(k)-temp < temp-pivksv) THEN
      pivksv = temp
      shfmax = shift
    ELSE
      pivksv = zero
      shfmax = (shift*piv(k) - shfmin*temp) / (piv(k)-temp)
    END IF
  ELSE
    ksave = k
    pivksv = temp
    shfmax = shift
  END IF
END IF
IF (shfmin <= 0.99D0*shfmax) GO TO 350
370 evalue = shfmin

!     Apply the inverse Householder transformations to D.

380 nm = n - 1
DO  k = nm, 1, -1
  kp = k + 1
  sum = zero
  DO  i = kp, n
    sum = sum + d(i) * h(i,k)
  END DO
  DO  i = kp, n
    d(i) = d(i) - sum * h(i,k)
  END DO
END DO

!     Return from the subroutine.

420 RETURN
END SUBROUTINE trstep

!%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% lagmax.f %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

SUBROUTINE lagmax(n, g, h, rho, d, v, vmax)

INTEGER, PARAMETER  :: dp = SELECTED_REAL_KIND(12, 60)
INTEGER, INTENT(IN)     :: n
REAL (dp), INTENT(IN)   :: g(:)
REAL (dp), INTENT(OUT)  :: h(:,:)
REAL (dp), INTENT(IN)   :: rho
REAL (dp), INTENT(OUT)  :: d(:)
REAL (dp), INTENT(OUT)  :: v(:)
REAL (dp), INTENT(OUT)  :: vmax

!     N is the number of variables of a quadratic objective function, Q say.
!     G is the gradient of Q at the origin.
!     H is the symmetric Hessian matrix of Q. Only the upper triangular and
!       diagonal parts need be set.
!     RHO is the trust region radius, and has to be positive.
!     D will be set to the calculated vector of variables.
!     The array V will be used for working space.
!     VMAX will be set to |Q(0)-Q(D)|.

!     Calculating the D that maximizes |Q(0)-Q(D)| subject to ||D|| <= RHO
!     requires of order N**3 operations, but sometimes it is adequate if
!     |Q(0)-Q(D)| is within about 0.9 of its greatest possible value. This
!     subroutine provides such a solution in only of order N**2 operations,
!     where the claim of accuracy has been tested by numerical experiments.

REAL (dp)  :: half = 0.5_dp, one = 1.0_dp, zero = 0.0_dp
REAL (dp)  :: dd, dhd, dlin, dsq, gd, gg, ghg, gnorm, halfrt, hmax, ratio
REAL (dp)  :: scale, sum, sumv, temp, tempa, tempb, tempc, tempd, tempv
REAL (dp)  :: vhg, vhv, vhw, vlin, vmu, vnorm, vsq, vv, wcos, whw, wsin, wsq
INTEGER    :: i, j, k

!     Preliminary calculations.

halfrt = SQRT(half)

!     Pick V such that ||HV|| / ||V|| is large.

hmax = zero
DO  i = 1, n
  sum = zero
  DO  j = 1, n
    h(j,i) = h(i,j)
    sum = sum + h(i,j) ** 2
  END DO
  IF (sum > hmax) THEN
    hmax = sum
    k = i
  END IF
END DO
DO  j = 1, n
  v(j) = h(k,j)
END DO

!     Set D to a vector in the subspace spanned by V and HV that maximizes
!     |(D,HD)|/(D,D), except that we set D=HV if V and HV are nearly parallel.
!     The vector that has the name D at label 60 used to be the vector W.

vsq = zero
vhv = zero
dsq = zero
DO  i = 1, n
  vsq = vsq + v(i) ** 2
  d(i) = DOT_PRODUCT( h(i,1:n), v(1:n) )
  vhv = vhv + v(i) * d(i)
  dsq = dsq + d(i) ** 2
END DO
IF (vhv*vhv <= 0.9999D0*dsq*vsq) THEN
  temp = vhv / vsq
  wsq = zero
  DO  i = 1, n
    d(i) = d(i) - temp * v(i)
    wsq = wsq + d(i) ** 2
  END DO
  whw = zero
  ratio = SQRT(wsq/vsq)
  DO  i = 1, n
    temp = DOT_PRODUCT( h(i,1:n), d(1:n) )
    whw = whw + temp * d(i)
    v(i) = ratio * v(i)
  END DO
  vhv = ratio * ratio * vhv
  vhw = ratio * wsq
  temp = half * (whw-vhv)
  temp = temp + SIGN(SQRT(temp**2+vhw**2),whw+vhv)
  DO  i = 1, n
    d(i) = vhw * v(i) + temp * d(i)
  END DO
END IF

!     We now turn our attention to the subspace spanned by G and D. A multiple
!     of the current D is returned if that choice seems to be adequate.

gg = zero
gd = zero
dd = zero
dhd = zero
DO  i = 1, n
  gg = gg + g(i) ** 2
  gd = gd + g(i) * d(i)
  dd = dd + d(i) ** 2
  sum = DOT_PRODUCT( h(i,1:n), d(1:n) )
  dhd = dhd + sum * d(i)
END DO
temp = gd / gg
vv = zero
scale = SIGN(rho/SQRT(dd),gd*dhd)
DO  i = 1, n
  v(i) = d(i) - temp * g(i)
  vv = vv + v(i) ** 2
  d(i) = scale * d(i)
END DO
gnorm = SQRT(gg)
IF (gnorm*dd <= 0.5D-2*rho*ABS(dhd) .OR. vv/dd <= 1.0D-4) THEN
  vmax = ABS(scale*(gd + half*scale*dhd))
  GO TO 170
END IF

!     G and V are now orthogonal in the subspace spanned by G and D.  Hence
!     we generate an orthonormal basis of this subspace such that (D,HV) is
!     negligible or zero, where D and V will be the basis vectors.

ghg = zero
vhg = zero
vhv = zero
DO  i = 1, n
  sum = DOT_PRODUCT( h(i,1:n), g(1:n) )
  sumv = DOT_PRODUCT( h(i,1:n), v(1:n) )
  ghg = ghg + sum * g(i)
  vhg = vhg + sumv * g(i)
  vhv = vhv + sumv * v(i)
END DO
vnorm = SQRT(vv)
ghg = ghg / gg
vhg = vhg / (vnorm*gnorm)
vhv = vhv / vv
IF (ABS(vhg) <= 0.01D0*MAX(ABS(ghg),ABS(vhv))) THEN
  vmu = ghg - vhv
  wcos = one
  wsin = zero
ELSE
  temp = half * (ghg-vhv)
  vmu = temp + SIGN(SQRT(temp**2+vhg**2),temp)
  temp = SQRT(vmu**2+vhg**2)
  wcos = vmu / temp
  wsin = vhg / temp
END IF
tempa = wcos / gnorm
tempb = wsin / vnorm
tempc = wcos / vnorm
tempd = wsin / gnorm
DO  i = 1, n
  d(i) = tempa * g(i) + tempb * v(i)
  v(i) = tempc * v(i) - tempd * g(i)
END DO

!     The final D is a multiple of the current D, V, D+V or D-V. We make the
!     choice from these possibilities that is optimal.

dlin = wcos * gnorm / rho
vlin = -wsin * gnorm / rho
tempa = ABS(dlin) + half * ABS(vmu+vhv)
tempb = ABS(vlin) + half * ABS(ghg-vmu)
tempc = halfrt * (ABS(dlin)+ABS(vlin)) + 0.25D0 * ABS(ghg+vhv)
IF (tempa >= tempb .AND. tempa >= tempc) THEN
  tempd = SIGN(rho,dlin*(vmu+vhv))
  tempv = zero
ELSE IF (tempb >= tempc) THEN
  tempd = zero
  tempv = SIGN(rho,vlin*(ghg-vmu))
ELSE
  tempd = SIGN(halfrt*rho,dlin*(ghg+vhv))
  tempv = SIGN(halfrt*rho,vlin*(ghg+vhv))
END IF
DO  i = 1, n
  d(i) = tempd * d(i) + tempv * v(i)
END DO
vmax = rho * rho * MAX(tempa,tempb,tempc)
170 RETURN
END SUBROUTINE lagmax


  SUBROUTINE calfun(nterms, a, f, iordr,yfit1,yfit2,yfit3,is,xms,yms,ecut2,lik_arr,nfel)
    IMPLICIT NONE
    INTEGER, PARAMETER  :: dp = SELECTED_REAL_KIND(12, 60)
    INTEGER, INTENT(IN)    :: nterms
    REAL     , INTENT(IN)  :: a(max_par)
    REAL (dp), INTENT(OUT) :: f
! additional parameters for emldetect use:
REAL                       :: xms(max_fit), yms(max_fit), lik_arr(max_ima)
integer                    :: iordr(max_par)
integer                    :: is, nfel, ierr 
real                       :: like0
real                       :: yfit1(max_events),yfit2(max_ima),yfit3,ecut2


! write(*,*) "k)"
       call funct(a,iordr,yfit1,yfit2,yfit3,is,nterms+nfel,xms,yms,ecut2)
       call like(like0,lik_arr,yfit1,yfit2) 
!       write(*,*) "in calfun:",like0

       f=dble(like0)

  END SUBROUTINE calfun


end module emlfit



  










